#!/usr/bin/env python
"""
Map panel of ncvue.

The panel allows plotting contour or mesh maps of georeferenced data.
Maps can be animated along the time axis.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   ncvMap

History
    * Written Dec 2020-Jan 2021 by Matthias Cuntz (mc (at) macu (dot) de)
    * Open new netcdf file, communicate via top widget,
      Jan 2021, Matthias Cuntz
    * Write coordinates and value on bottom of plotting canvas,
      May 2021, Matthias Cuntz
    * Larger pad for colorbar, Jun 2021, Matthias Cuntz
    * Work with files without an unlimited (time) dimension (set_tstep),
      Oct 2021, Matthias Cuntz

"""
from __future__ import absolute_import, division, print_function
import sys
import tkinter as tk
try:
    import tkinter.ttk as ttk
except Exception:
    print('Using the themed widget set introduced in Tk 8.5.')
    sys.exit()
from tkinter import filedialog
import os
import numpy as np
import netCDF4 as nc
from .ncvutils   import add_cyclic, clone_ncvmain, format_coord_map
from .ncvutils   import set_axis_label, set_miss, vardim2var
from .ncvmethods import analyse_netcdf, get_slice_miss, get_miss
from .ncvmethods import set_dim_lon, set_dim_lat, set_dim_var
from .ncvwidgets import add_checkbutton, add_combobox, add_entry, add_imagemenu
from .ncvwidgets import add_menu, add_scale, add_spinbox, add_tooltip
import matplotlib as mpl
# mpl.use('TkAgg')
from matplotlib import pyplot as plt
# plt.style.use('seaborn-darkgrid')
plt.style.use('seaborn-dark')
# plt.style.use('fast')
import cartopy.crs as ccrs


__all__ = ['ncvMap']


class ncvMap(ttk.Frame):
    """
    Panel for maps.

    Sets up the layout with the figure canvas, variable selectors, dimension
    spinboxes, and options in __init__.

    Contains various commands that manage what will be drawn or redrawn if
    something is selected, changed, checked, etc.

    """

    #
    # Panel setup
    #

    def __init__(self, master, **kwargs):
        from functools import partial
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
        from matplotlib.figure import Figure
        from matplotlib import animation

        super().__init__(master, **kwargs)

        self.name   = 'Map'
        self.master = master
        self.top    = master.top
        # copy for ease of use
        self.fi     = self.top.fi
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols

        # unlimited dimension control
        self.iunlim = -1  # index of dunlim in dimensions of current var
        self.nunlim = 0   # length of dunlim of current plot variable

        # new window
        self.rowwin = ttk.Frame(self)
        self.rowwin.pack(side=tk.TOP, fill=tk.X)
        self.newfile = ttk.Button(self.rowwin, text="Open File",
                                  command=self.newnetcdf)
        self.newfile.pack(side=tk.LEFT)
        self.newfiletip = add_tooltip(self.newfile, 'Open a new netcdf file')
        spacew = ttk.Label(self.rowwin, text=" "*3)
        spacew.pack(side=tk.LEFT)
        time_label1 = ttk.Label(self.rowwin, text='Time: ')
        time_label1.pack(side=tk.LEFT)
        self.timelbl = tk.StringVar()
        self.timelbl.set('')
        time_label2 = ttk.Label(self.rowwin, textvariable=self.timelbl)
        time_label2.pack(side=tk.LEFT)
        self.newwin = ttk.Button(
            self.rowwin, text="New Window",
            command=partial(clone_ncvmain, self.master))
        self.newwin.pack(side=tk.RIGHT)
        self.newwintip = add_tooltip(
            self.newwin, 'Open secondary ncvue window')

        # plotting canvas
        self.figure = Figure(facecolor="white", figsize=(1, 1))
        # self.axes   = self.figure.add_subplot(111)
        self.axes   = self.figure.add_subplot(111,
                                              projection=ccrs.PlateCarree())
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.draw()
        # pack
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        # grid instead of pack - does not work
        # self.canvas.get_tk_widget().grid(column=0, row=0,
        #     sticky=(tk.N, tk.S, tk.E, tk.W))
        # self.canvas.get_tk_widget().columnconfigure(0, weight=1)
        # self.canvas.get_tk_widget().rowconfigure(0, weight=1)

        # matplotlib toolbar
        self.toolbar = NavigationToolbar2Tk(self.canvas, self)
        self.toolbar.update()
        self.toolbar.pack(side=tk.TOP, fill=tk.X)

        # selections and options
        columns = [''] + self.cols

        allcmaps = plt.colormaps()
        self.cmaps  = [ i for i in allcmaps if not i.endswith('_r') ]
        self.cmaps.sort()
        # self.imaps  = [ tk.PhotoImage(file=os.path.dirname(__file__) +
        #                               '/images/' + i + '.png')
        #                 for i in self.cmaps ]
        bundle_dir = getattr(sys, '_MEIPASS',
                             os.path.abspath(os.path.dirname(__file__)))
        self.imaps  = [ tk.PhotoImage(file=bundle_dir +
                                      '/images/' + i + '.png')
                        for i in self.cmaps ]

        # only projections with keyword: central_longitude
        self.projs = ['AlbersEqualArea', 'AzimuthalEquidistant', 'EckertI',
                      'EckertII', 'EckertIII', 'EckertIV', 'EckertV',
                      'EckertVI', 'EqualEarth', 'EquidistantConic',
                      'InterruptedGoodeHomolosine',
                      'LambertAzimuthalEqualArea', 'LambertConformal',
                      'LambertCylindrical', 'Mercator', 'Miller', 'Mollweide',
                      'NorthPolarStereo', 'PlateCarree', 'Robinson',
                      'Sinusoidal', 'SouthPolarStereo', 'Stereographic']
        self.iprojs = [ccrs.AlbersEqualArea, ccrs.AzimuthalEquidistant,
                       ccrs.EckertI, ccrs.EckertII, ccrs.EckertIII,
                       ccrs.EckertIV, ccrs.EckertV, ccrs.EckertVI,
                       ccrs.EqualEarth, ccrs.EquidistantConic,
                       ccrs.InterruptedGoodeHomolosine,
                       ccrs.LambertAzimuthalEqualArea, ccrs.LambertConformal,
                       ccrs.LambertCylindrical, ccrs.Mercator, ccrs.Miller,
                       ccrs.Mollweide, ccrs.NorthPolarStereo, ccrs.PlateCarree,
                       ccrs.Robinson, ccrs.Sinusoidal, ccrs.SouthPolarStereo,
                       ccrs.Stereographic]

        # 1. row
        # controls
        self.rowt = ttk.Frame(self)
        self.rowt.pack(side=tk.TOP, fill=tk.X)
        ntime = 1
        self.tsteplbl, self.tstepval, self.tstep, self.tsteptip = add_scale(
            self.rowt, label="step", ini=0, from_=0, to=ntime,
            length=100, orient=tk.HORIZONTAL, command=self.tstep_t,
            tooltip="Slide to go to time step")
        spacet = ttk.Label(self.rowt, text=" "*1)
        spacet.pack(side=tk.LEFT)
        # first t
        self.first_t = ttk.Button(self.rowt, text="|<<", width=3,
                                  command=self.first_t)
        self.first_t.pack(side=tk.LEFT)
        self.first_ttip = add_tooltip(self.first_t, 'First time step')
        # previous t
        self.prev_t = ttk.Button(self.rowt, text="|<", width=2,
                                 command=self.prev_t)
        self.prev_t.pack(side=tk.LEFT)
        self.prev_ttip = add_tooltip(self.prev_t, 'Previous time step')
        # run t backwards
        self.prun_t = ttk.Button(self.rowt, text="<", width=1,
                                 command=self.prun_t)
        self.prun_t.pack(side=tk.LEFT)
        self.prun_ttip = add_tooltip(self.prun_t, 'Run backwards')
        # pause t
        self.pause_t = ttk.Button(self.rowt, text="||", width=1,
                                  command=self.pause_t)
        self.pause_t.pack(side=tk.LEFT)
        self.pause_ttip = add_tooltip(self.pause_t, 'Pause/Stop')
        # run t forward
        self.nrun_t = ttk.Button(self.rowt, text=">", width=1,
                                 command=self.nrun_t)
        self.nrun_t.pack(side=tk.LEFT)
        self.nrun_ttip = add_tooltip(self.nrun_t, 'Run forwards')
        # next t
        self.next_t = ttk.Button(self.rowt, text=">|", width=2,
                                 command=self.next_t)
        self.next_t.pack(side=tk.LEFT)
        self.next_ttip = add_tooltip(self.next_t, 'Next time step')
        # last t
        self.last_t = ttk.Button(self.rowt, text=">>|", width=3,
                                 command=self.last_t)
        self.last_t.pack(side=tk.LEFT)
        self.last_ttip = add_tooltip(self.last_t, 'Last time step')
        # repeat
        spacer = ttk.Label(self.rowt, text=" "*1)
        spacer.pack(side=tk.LEFT)
        reps = ['once', 'repeat', 'reflect']
        tstr  = "Run time steps once, repeat from start when at end,"
        tstr += " or continue running backwards when at end"
        self.repeatlbl, self.repeat, self.repeattip = add_combobox(
            self.rowt, label="repeat", values=reps, width=5,
            command=self.repeat_t, tooltip=tstr)
        self.repeat.set('repeat')
        self.last_t.pack(side=tk.LEFT)
        # delay
        spaced = ttk.Label(self.rowt, text=" "*1)
        spaced.pack(side=tk.LEFT)
        tstr = "Delay run between time steps from 1 to 1000 ms"
        self.delaylbl, self.delayval, self.delay, self.delaytip = add_scale(
            self.rowt, label="delay (ms)", ini=1, from_=1, to=1000,
            length=100, orient=tk.HORIZONTAL, command=self.delay_t,
            tooltip=tstr)

        # 2. row
        # variable-axis selection
        self.rowvv = ttk.Frame(self)
        self.rowvv.pack(side=tk.TOP, fill=tk.X)
        self.blockv = ttk.Frame(self.rowvv)
        self.blockv.pack(side=tk.LEFT)
        self.rowv = ttk.Frame(self.blockv)
        self.rowv.pack(side=tk.TOP, fill=tk.X)
        self.vlbl = tk.StringVar()
        self.vlbl.set("var")
        vlab = ttk.Label(self.rowv, textvariable=self.vlbl)
        vlab.pack(side=tk.LEFT)
        self.bprev_v = ttk.Button(self.rowv, text="<", width=1,
                                  command=self.prev_v)
        self.bprev_v.pack(side=tk.LEFT)
        self.bprev_vtip = add_tooltip(self.bprev_v, 'Previous variable')
        self.bnext_v = ttk.Button(self.rowv, text=">", width=1,
                                  command=self.next_v)
        self.bnext_v.pack(side=tk.LEFT)
        self.bnext_vtip = add_tooltip(self.bnext_v, 'Next variable')
        self.v = ttk.Combobox(self.rowv, values=columns, width=25)
        self.v.bind("<<ComboboxSelected>>", self.selected_v)
        self.v.pack(side=tk.LEFT)
        self.vtip = add_tooltip(self.v, 'Choose variable')
        self.trans_vlbl, self.trans_v, self.trans_vtip = add_checkbutton(
            self.rowv, label="transpose var", value=False,
            command=self.checked,
            tooltip="Transpose array, i.e. exchanging lat and lon")
        spacev = ttk.Label(self.rowv, text=" "*1)
        spacev.pack(side=tk.LEFT)
        self.vminlbl, self.vmin, self.vmintip = add_entry(
            self.rowv, label="vmin", text=0, width=11, command=self.entered_v,
            tooltip="Minimal display value")
        self.vmaxlbl, self.vmax, self.vmaxtip = add_entry(
            self.rowv, label="vmax", text=1, width=11, command=self.entered_v,
            tooltip="Maximal display value")
        tstr  = "If checked, determine vmin/vmax from all fields,\n"
        tstr += "otherwise from 50 random fields"
        self.valllbl, self.vall, self.valltip = add_checkbutton(
            self.rowv, label="all", value=False, command=self.checked_all,
            tooltip=tstr)
        # levels var
        self.rowvd = ttk.Frame(self.blockv)
        self.rowvd.pack(side=tk.TOP, fill=tk.X)
        self.vdlblval = []
        self.vdlbl    = []
        self.vdval    = []
        self.vd       = []
        self.vdtip    = []
        for i in range(self.maxdim):
            vdlblval, vdlbl, vdval, vd, vdtip = add_spinbox(
                self.rowvd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_v, state=tk.DISABLED, tooltip="None")
            self.vdlblval.append(vdlblval)
            self.vdlbl.append(vdlbl)
            self.vdval.append(vdval)
            self.vd.append(vd)
            self.vdtip.append(vdtip)

        # 3. row
        # lon-axis selection
        self.rowll = ttk.Frame(self)
        self.rowll.pack(side=tk.TOP, fill=tk.X)
        self.blocklon = ttk.Frame(self.rowll)
        self.blocklon.pack(side=tk.LEFT)
        self.rowlon = ttk.Frame(self.blocklon)
        self.rowlon.pack(side=tk.TOP, fill=tk.X)
        self.lonlbl, self.lon, self.lontip = add_combobox(
            self.rowlon, label="lon", values=columns,
            command=self.selected_lon,
            tooltip="Longitude variable.\nSet 'empty' for matrix plot.")
        self.inv_lonlbl, self.inv_lon, self.inv_lontip = add_checkbutton(
            self.rowlon, label="invert lon", value=False,
            command=self.checked,
            tooltip="Invert longitudes")
        self.shift_lonlbl, self.shift_lon, self.shift_lontip = add_checkbutton(
            self.rowlon, label="shift lon/2", value=False,
            command=self.checked,
            tooltip="Roll longitudes by half its size")
        self.rowlond = ttk.Frame(self.blocklon)
        self.rowlond.pack(side=tk.TOP, fill=tk.X)
        self.londlblval = []
        self.londlbl    = []
        self.londval    = []
        self.lond       = []
        self.londtip    = []
        for i in range(self.maxdim):
            londlblval, londlbl, londval, lond, londtip = add_spinbox(
                self.rowlond, label=str(i), values=(0,), wrap=True,
                command=self.spinned_lon, state=tk.DISABLED, tooltip="None")
            self.londlblval.append(londlblval)
            self.londlbl.append(londlbl)
            self.londval.append(londval)
            self.lond.append(lond)
            self.londtip.append(londtip)
        # lat-axis selection
        spacex = ttk.Label(self.rowll, text=" "*3)
        spacex.pack(side=tk.LEFT)
        self.blocklat = ttk.Frame(self.rowll)
        self.blocklat.pack(side=tk.LEFT)
        self.rowlat = ttk.Frame(self.blocklat)
        self.rowlat.pack(side=tk.TOP, fill=tk.X)
        self.latlbl, self.lat, self.lattip = add_combobox(
            self.rowlat, label="lat", values=columns,
            command=self.selected_lat,
            tooltip="Longitude variable.\nSet 'empty' for matrix plot.")
        self.inv_latlbl, self.inv_lat, self.inv_lattip = add_checkbutton(
            self.rowlat, label="invert lat", value=False, command=self.checked,
            tooltip="Invert longitudes")
        self.rowlatd = ttk.Frame(self.blocklat)
        self.rowlatd.pack(side=tk.TOP, fill=tk.X)
        self.latdlblval = []
        self.latdlbl    = []
        self.latdval    = []
        self.latd       = []
        self.latdtip    = []
        for i in range(self.maxdim):
            latdlblval, latdlbl, latdval, latd, latdtip = add_spinbox(
                self.rowlatd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_lat, state=tk.DISABLED, tooltip="None")
            self.latdlblval.append(latdlblval)
            self.latdlbl.append(latdlbl)
            self.latdval.append(latdval)
            self.latd.append(latd)
            self.latdtip.append(latdtip)

        # 4. row
        # options
        self.rowcmap = ttk.Frame(self)
        self.rowcmap.pack(side=tk.TOP, fill=tk.X)
        self.cmaplbl, self.cmap, self.cmaptip = add_imagemenu(
            self.rowcmap, label="cmap", values=self.cmaps,
            images=self.imaps, command=self.selected_cmap,
            tooltip="Choose colormap")
        self.cmap['text']  = 'RdYlBu'
        self.cmap['image'] = self.imaps[self.cmaps.index('RdYlBu')]
        self.rev_cmaplbl, self.rev_cmap, self.rev_cmaptip = add_checkbutton(
            self.rowcmap, label="reverse cmap", value=False,
            command=self.checked,
            tooltip="Reverse colormap")
        self.meshlbl, self.mesh, self.meshtip = add_checkbutton(
            self.rowcmap, label="mesh", value=True,
            command=self.checked,
            tooltip="Pseudocolor plot if checked, plot contours if unchecked")
        self.igloballbl, self.iglobal, self.iglobaltip = add_checkbutton(
            self.rowcmap, label="global", value=False,
            command=self.checked,
            tooltip="Assume global extent")
        self.coastlbl, self.coast, self.coasttip = add_checkbutton(
            self.rowcmap, label="coast", value=True,
            command=self.checked,
            tooltip="Draw continental coast lines")
        self.gridlbl, self.grid, self.gridtip = add_checkbutton(
            self.rowcmap, label="grid", value=False,
            command=self.checked,
            tooltip="Draw major grid lines")

        # 7. row
        # projections
        self.rowproj = ttk.Frame(self)
        self.rowproj.pack(side=tk.TOP, fill=tk.X)
        self.projlbl, self.proj, self.projtip = add_menu(
            self.rowproj, label="projection", values=self.projs,
            command=self.selected_proj, width=26,
            tooltip="Choose projection")
        self.proj['text'] = 'PlateCarree'
        tstr  = "Central longitude of projection.\n"
        tstr += "Determined from longitude variable if None."
        self.clonlbl, self.clon, self.clontip = add_entry(
            self.rowproj, label="central lon", text='None', width=4,
            command=self.entered_clon, tooltip=tstr)

        # set lat/lon
        if self.latvar:
            self.lat.set(self.latvar)
            self.inv_lat.set(0)
            set_dim_lat(self)
        if self.lonvar:
            self.lon.set(self.lonvar)
            self.inv_lon.set(0)
            self.shift_lon.set(0)
            set_dim_lon(self)

        # set global
        x = self.lon.get()
        if (x != ''):
            vx = vardim2var(x)
            xx = self.fi.variables[vx]
            xx = get_slice_miss(self, self.lond, xx)
            if np.any(np.isfinite(xx)):
                xx = (xx + 360.) % 360.
            if (xx.max() - xx.min()) > 150.:
                self.iglobal.set(1)
            else:
                self.iglobal.set(0)

        # animation
        rep = self.repeat.get()
        if rep == 'repeat':
            irepeat = True
        else:
            irepeat = False
        self.anim_first   = True   # True: stops in self.update at first call
        self.anim_running = True   # True/False: animation running or not
        self.anim_inc     = 1      # 1/-1: forward or backward run
        self.anim = animation.FuncAnimation(self.figure, self.update,
                                            init_func=self.redraw,
                                            interval=self.delayval.get(),
                                            repeat=irepeat)

    #
    # Bindings
    #

    def checked(self):
        """
        Command called if any checkbutton was checked or unchecked.

        Redraws plot.
        """
        self.redraw()

    def checked_all(self):
        """
        Command called if any checkbutton 'all' for vmin/vmax was checked or
        unchecked.

        Resets vmin/vmax, redraws plot.
        """
        vmin, vmax = self.get_vminmax()
        self.vmin.set(vmin)
        self.vmax.set(vmax)
        self.redraw()

    def delay_t(self, delay):
        """
        Command called if delay scale was changed.

        `delay` is the chosen value on the scale slider.
        """
        self.anim.event_source.interval = int(float(delay))

    def entered_clon(self, event):
        """
        Command called if values central longitude was entered.

        Triggering `event` was bound to entry.

        Redraws plot.
        """
        self.redraw()

    def entered_v(self, event):
        """
        Command called if values for `vmin`/`vmax` were entered.

        Triggering `event` was bound to entry.

        Redraws plot.
        """
        self.redraw()

    def first_t(self):
        """
        Command called if first frame button was pressed.
        """
        it = 0
        self.set_tstep(it)
        self.update(it, isframe=True)

    def last_t(self):
        """
        Command called if last frame button was pressed.
        """
        it = self.nunlim - 1
        self.set_tstep(it)
        self.update(it, isframe=True)

    def newnetcdf(self):
        """
        Open a new netcdf file and connect it to top.
        """
        # stop animation if running
        if self.anim_running:
            self.anim.event_source.stop()
            self.anim_running = False
        # get new netcdf file name
        ncfile = filedialog.askopenfilename(
            parent=self, title='Choose netcdf file', multiple=False)
        if ncfile:
            # close old netcdf file
            if self.top.fi:
                self.top.fi.close()
            # reset empty defaults of top
            self.top.dunlim = ''      # name of unlimited dimension
            self.top.time   = None    # datetime variable
            self.top.tname  = ''      # datetime variable name
            self.top.tvar   = ''      # datetime variable name in netcdf
            self.top.dtime  = None    # decimal year
            self.top.latvar = ''      # name of latitude variable
            self.top.lonvar = ''      # name of longitude variable
            self.top.latdim = ''      # name of latitude dimension
            self.top.londim = ''      # name of longitude dimension
            self.top.maxdim = 0       # maximum num of dims of all variables
            self.top.cols   = []      # variable list
            # open new netcdf file
            self.top.fi = nc.Dataset(ncfile, 'r')
            analyse_netcdf(self.top)
            # reset panel
            self.reinit()
            self.redraw()

    def nrun_t(self):
        """
        Command called if forward run button was pressed.
        """
        if not self.anim_running:
            self.anim_inc = 1
            self.anim.event_source.start()
            self.anim_running = True

    def next_t(self):
        """
        Command called if next frame button was pressed.
        """
        it = int(self.vdval[self.iunlim].get())
        if it < self.nunlim-1:
            it += 1
            self.set_tstep(it)
            self.update(it, isframe=True)
        else:
            rep = self.repeat.get()
            if rep != 'once':
                if rep == 'repeat':
                    it = 0
                else:  # reflect
                    it -= 1
                self.set_tstep(it)
                self.update(it, isframe=True)

    def next_v(self):
        """
        Command called if next button for the plotting variable was pressed.

        Resets `vmin`/`vmax` and variable-dimensions. Redraws plot.
        """
        v = self.v.get()
        cols = self.v["values"]
        idx  = cols.index(v)
        idx += 1
        if idx < len(cols):
            self.v.set(cols[idx])
            self.set_unlim(cols[idx])
            self.tstep['to'] = self.nunlim - 1
            self.set_tstep(0)
            vmin, vmax = self.get_vminmax()
            self.vmin.set(vmin)
            self.vmax.set(vmax)
            set_dim_var(self)
            self.redraw()

    def pause_t(self):
        """
        Command called if pause button was pressed.
        """
        if self.anim_running:
            self.anim.event_source.stop()
            self.anim_running = False

    def prev_t(self):
        """
        Command called if previous frame button was pressed.
        """
        it = int(self.vdval[self.iunlim].get())
        if it > 0:
            it -= 1
            self.set_tstep(it)
            self.update(it, isframe=True)
        else:
            rep = self.repeat.get()
            if rep != 'once':
                if rep == 'repeat':
                    it = self.nunlim - 1
                else:  # reflect
                    it += 1
                self.set_tstep(it)
                self.update(it, isframe=True)

    def prev_v(self):
        """
        Command called if previous button for the plotting variable was
        pressed.

        Resets `vmin`/`vmax` and v-dimensions. Redraws plot.
        """
        v = self.v.get()
        cols = self.v["values"]
        idx  = cols.index(v)
        idx -= 1
        if idx > 0:
            self.v.set(cols[idx])
            self.set_unlim(cols[idx])
            self.tstep['to'] = self.nunlim - 1
            self.set_tstep(0)
            vmin, vmax = self.get_vminmax()
            self.vmin.set(vmin)
            self.vmax.set(vmax)
            set_dim_var(self)
            self.redraw()

    def prun_t(self):
        """
        Command called if run backwards button was pressed.
        """
        if not self.anim_running:
            self.anim_inc = -1
            self.anim.event_source.start()
            self.anim_running = True

    def repeat_t(self, event):
        """
        Command called if repeat option was chosen with combobox.

        Triggering `event` was bound to the combobox.
        """
        rep = self.repeat.get()
        if rep == 'once':
            irepeat = False
        else:
            # need not to stop also for reflect
            irepeat = True
        self.anim.repeat = irepeat

    def selected_cmap(self, value):
        """
        Command called if cmap was chosen from menu.

        `value` is the chosen colormap.

        Sets text and image on the menubutton.
        """
        self.cmap['text']  = value
        self.cmap['image'] = self.imaps[self.cmaps.index(value)]
        self.redraw()

    def selected_lat(self, event):
        """
        Command called if latitude-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `lat` options and dimensions. Redraws plot.
        """
        self.inv_lat.set(0)
        set_dim_lat(self)
        self.redraw()

    def selected_lon(self, event):
        """
        Command called if x-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `x` options and dimensions. Redraws plot.
        """
        self.inv_lon.set(0)
        self.shift_lon.set(0)
        set_dim_lon(self)
        self.redraw()

    def selected_proj(self, value):
        """
        Command called if projection was chosen from menu.

        `value` is the chosen projection.

        Sets text on the menubutton.
        """
        self.proj['text'] = value
        self.redraw()

    def selected_v(self, event):
        """
        Command called if plotting variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `vmin`/`vmax` and variable-dimensions. Redraws plot.
        """
        v = self.v.get()
        self.set_unlim(v)
        self.tstep['to'] = self.nunlim - 1
        self.set_tstep(0)
        vmin, vmax = self.get_vminmax()
        self.vmin.set(vmin)
        self.vmax.set(vmax)
        set_dim_var(self)
        self.redraw()

    def spinned_lon(self, event=None):
        """
        Command called if spinbox of x-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        self.redraw()

    def spinned_lat(self, event=None):
        """
        Command called if spinbox of latitude-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        self.redraw()

    def spinned_v(self, event=None):
        """
        Command called if spinbox of variable-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        try:
            it = int(self.vdval[self.iunlim].get())
            self.set_tstep(it)
        except ValueError:  # mean, std, etc.
            pass
        self.redraw()

    def tstep_t(self, step):
        """
        Command called if tstep was changed.

        `step` is the chosen value on the scale slider.
        """
        it = int(float(step))
        self.set_tstep(it)
        self.update(it, isframe=True)

    #
    # Methods
    #

    def get_vminmax(self):
        from numpy.random import default_rng
        v = self.v.get()
        if (v != ''):
            vz = vardim2var(v)
            if vz == self.tname:
                return (0, 1)
            vv = self.fi.variables[vz]
            imiss = get_miss(self, vv)
            iall  = self.vall.get()
            if iall or (np.sum(vv.shape[:-2]) < 50):
                vv   = set_miss(imiss, vv)
                vmin = np.nanmin(vv)
                vmax = np.nanmax(vv)
            else:
                rng = default_rng()
                vmin = np.inf
                vmax = -np.inf
                for nn in range(50):
                    ss = []
                    for i in range(vv.ndim):
                        if i < vv.ndim-2:
                            idim = rng.integers(0, vv.shape[i])
                            s = slice(idim, idim+1)
                        else:
                            s = slice(0, vv.shape[i])
                        ss.append(s)
                    ivv   = vv[tuple(ss)]
                    ivv   = set_miss(imiss, ivv)
                    ivmin = np.nanmin(ivv)
                    ivmax = np.nanmax(ivv)
                    vmin  = min(vmin, ivmin)
                    vmax  = max(vmax, ivmax)
            return (vmin, vmax)
        else:
            return (0, 1)

    def reinit(self):
        """
        Reinitialise the panel from top.
        """
        # reinit from top
        self.fi     = self.top.fi
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols
        self.iunlim = -1
        self.nunlim = 0
        # reset dimensions
        for ll in self.vdlbl:
            ll.destroy()
        for ll in self.vd:
            ll.destroy()
        self.vdlblval = []
        self.vdlbl    = []
        self.vdval    = []
        self.vd       = []
        self.vdtip    = []
        for i in range(self.maxdim):
            vdlblval, vdlbl, vdval, vd, vdtip = add_spinbox(
                self.rowvd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_v, state=tk.DISABLED, tooltip="None")
            self.vdlblval.append(vdlblval)
            self.vdlbl.append(vdlbl)
            self.vdval.append(vdval)
            self.vd.append(vd)
            self.vdtip.append(vdtip)
        for ll in self.latdlbl:
            ll.destroy()
        for ll in self.latd:
            ll.destroy()
        self.latdlblval = []
        self.latdlbl    = []
        self.latdval    = []
        self.latd       = []
        self.latdtip    = []
        for i in range(self.maxdim):
            latdlblval, latdlbl, latdval, latd, latdtip = add_spinbox(
                self.rowlatd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_lat, state=tk.DISABLED, tooltip="None")
            self.latdlblval.append(latdlblval)
            self.latdlbl.append(latdlbl)
            self.latdval.append(latdval)
            self.latd.append(latd)
            self.latdtip.append(latdtip)
        for ll in self.londlbl:
            ll.destroy()
        for ll in self.lond:
            ll.destroy()
        self.londlblval = []
        self.londlbl    = []
        self.londval    = []
        self.lond       = []
        self.londtip    = []
        for i in range(self.maxdim):
            londlblval, londlbl, londval, lond, londtip = add_spinbox(
                self.rowlond, label=str(i), values=(0,), wrap=True,
                command=self.spinned_lon, state=tk.DISABLED, tooltip="None")
            self.londlblval.append(londlblval)
            self.londlbl.append(londlbl)
            self.londval.append(londval)
            self.lond.append(lond)
            self.londtip.append(londtip)
        # set time step
        self.tstep['to'] = 0
        self.tstepval.set(0)
        self.repeat.set('repeat')
        # set variables
        columns = [''] + self.cols
        self.v['values'] = columns
        self.v.set(columns[0])
        self.vmin.set('None')
        self.vmax.set('None')
        # set lat/lon
        self.lon['values'] = columns
        self.lon.set(columns[0])
        self.lat['values'] = columns
        self.lat.set(columns[0])
        if self.latvar:
            self.lat.set(self.latvar)
            self.inv_lat.set(0)
            set_dim_lat(self)
        if self.lonvar:
            self.lon.set(self.lonvar)
            self.inv_lon.set(0)
            self.shift_lon.set(0)
            set_dim_lon(self)
        x = self.lon.get()
        if (x != ''):
            vx = vardim2var(x)
            xx = self.fi.variables[vx]
            xx = get_slice_miss(self, self.lond, xx)
            xx = xx + 360.
            if (xx.max() - xx.min()) > 150.:
                self.iglobal.set(1)
            else:
                self.iglobal.set(0)

    def set_tstep(self, it):
        """
        Make all steps when changing time step.

        `it` (int) is the time step.

        Sets the time dimension spinbox, sets the time step scale,
        write the time on top.
        """
        if self.dunlim:
            self.vdval[self.iunlim].set(it)
            self.tstepval.set(it)
            try:
                self.timelbl.set(np.around(self.time[it], 4))
            except TypeError:
                self.timelbl.set(self.time[it])

    def set_unlim(self, v):
        """
        Set index and length of unlimited dimension of variable `v`.

        `v` (str) is the variable name as in the selection comboboxes, i.e.
        `var = vardim2var(self.fi.variables[v)]`.

        Sets `self.nunlim` to the length of the unlimited dimension and
        `self.iunlim` to the index in variable.dimensions if
        `self.dunlim ~= ''` and `self.dunlim` in var.dimensions.

        Takes `self.iunlim=0` and `self.nunlim=variable.shape[0]` if
        self.dunlim == ''` or `self.dunlim` not in var.dimensions.
        """
        vz = vardim2var(v)
        if vz == self.tname:
            self.iunlim = 0
            self.nunlim = self.time.size
        else:
            if self.dunlim:
                if self.dunlim in self.fi.variables[vz].dimensions:
                    self.iunlim = (
                        self.fi.variables[vz].dimensions.index(self.dunlim))
                else:
                    self.iunlim = 0
            else:
                self.iunlim = 0
            if self.fi.variables[vz].ndim > 0:
                self.nunlim = self.fi.variables[vz].shape[self.iunlim]
            else:
                self.nunlim = 0

    #
    # Plotting
    #

    def redraw(self):
        """
        Redraws the plot.

        Reads `lon`, `lat`, `variable` names, the current settings of
        their dimension spinboxes, as well as all other plotting options.
        Then redraws the plot.
        """
        # stop animation
        self.anim.event_source.stop()
        self.anim_running = False
        # get all states
        # rowv
        v = self.v.get()
        trans_v = self.trans_v.get()
        vmin = self.vmin.get()
        if vmin == 'None':
            vmin = None
        else:
            try:
                vmin = float(vmin)
            except ValueError:
                vmin = None
        vmax = self.vmax.get()
        if vmax == 'None':
            vmax = None
        else:
            try:
                vmax = float(vmax)
            except ValueError:
                vmax = None
        # rowll
        x = self.lon.get()
        y = self.lat.get()
        inv_lon   = self.inv_lon.get()
        inv_lat   = self.inv_lat.get()
        shift_lon = self.shift_lon.get()
        # rowcmap
        cmap     = self.cmap['text']
        rev_cmap = self.rev_cmap.get()
        mesh     = self.mesh.get()
        self.iiglobal = self.iglobal.get()
        grid     = self.grid.get()
        coast    = self.coast.get()
        proj     = self.proj['text']
        self.iproj = self.iprojs[self.projs.index(proj)]
        clon     = self.clon.get()
        # set x, y, axes labels
        vx = 'None'
        vy = 'None'
        vz = 'None'
        if (v != ''):
            # variable
            vz = vardim2var(v)
            if vz == self.tname:
                # should throw an error later
                if mesh:
                    vv = self.dtime
                    vlab = 'Year'
                else:
                    vv = self.time
                    vlab = 'Date'
            else:
                vv = self.fi.variables[vz]
                vlab = set_axis_label(vv)
            vv = get_slice_miss(self, self.vd, vv)
            if trans_v:
                vv = vv.T
            if shift_lon:
                vv = np.roll(vv, vv.shape[1]//2, axis=1)
        else:
            vlab = ''
        if (y != ''):
            # y axis
            vy = vardim2var(y)
            if vy == self.tname:
                if mesh:
                    yy = self.dtime
                    ylab = 'Year'
                else:
                    yy = self.time
                    ylab = 'Date'
            else:
                yy   = self.fi.variables[vy]
                ylab = set_axis_label(yy)
            yy = get_slice_miss(self, self.latd, yy)
        else:
            ylab = ''
        if (x != ''):
            # x axis
            vx = vardim2var(x)
            if vx == self.tname:
                if mesh:
                    xx = self.dtime
                    xlab = 'Year'
                else:
                    xx = self.time
                    xlab = 'Date'
            else:
                xx   = self.fi.variables[vx]
                xlab = set_axis_label(xx)
            xx = get_slice_miss(self, self.lond, xx)
            # set central longitude of projection
            # make in 0-360, otherwise always 0 if -180 to 180
            if np.any(np.isfinite(xx)):
                xx360 = (xx + 360.) % 360.
            else:
                xx360 = xx
            if xx.size > 1:
                if xx.ndim > 1:
                    x0 = xx[:, 0].mean()
                    if self.iiglobal:
                        # -2 instead of -1 to avoid possible cyclic longitude
                        x1 = xx[:, -2].mean()
                    else:
                        x1 = xx[:, -1].mean()
                else:
                    x0 = xx[0]
                    if self.iiglobal:
                        x1 = xx[-2]
                    else:
                        x1 = xx[-1]
                self.ixxmean = 0.5 * (x1 + x0)
                if self.iiglobal:
                    # round it to next 180 degrees to get 0 or 180
                    self.ixxmean = np.around(self.ixxmean / 180., 0) * 180.
            else:
                self.ixxmean = xx360[0]
            # seems to work better if central lon in projection
            # is set from -180 to 180 even if lon is given in 0-360
            if self.ixxmean > 180.:
                self.ixxmean -= 360.
        else:
            xlab = ''
            self.ixxmean = 0.
        if clon != 'None':
            self.iclon = float(clon)
        else:
            self.iclon = self.ixxmean
        # plot options
        if rev_cmap:
            cmap = cmap + '_r'
        # Clear figure instead of axes because colorbar is on figure
        self.figure.clear()
        # Have to add axes again.
        self.axes = self.figure.add_subplot(
            111, projection=self.iproj(central_longitude=self.iclon))
        # plot only if variable given
        if (v != ''):
            if vv.ndim < 2:
                estr  = 'Map: var (' + vz + ') is not 2-dimensional:'
                print(estr, vv.shape)
                return
            # set x and y to index if not selected
            if (x == ''):
                nx = vv.shape[1]
                xx = -180. + np.arange(nx) / float(nx) * 360.
                xx += 0.5 * (xx[1] - xx[0])
                xlab = ''
            if (y == ''):
                ny = vv.shape[0]
                yy = -90. + np.arange(ny) / float(ny) * 180.
                yy += 0.5 * (yy[1] - yy[0])
                ylab = ''
            # plot
            # cc = self.axes.imshow(vv[:, ::-1].T, aspect='auto', cmap=cmap,
            #                       interpolation='none')
            # cc = self.axes.matshow(vv[:, ::-1].T, aspect='auto', cmap=cmap,
            #                        interpolation='none')
            extend = 'neither'
            if vmin is not None:
                vv = np.maximum(vv, vmin)
                if vmax is None:
                    extend = 'min'
                else:
                    extend = 'both'
            if vmax is not None:
                vv = np.minimum(vv, vmax)
                if vmin is None:
                    extend = 'max'
                else:
                    extend = 'both'
            if (xx.ndim == 1) and (yy.ndim == 1):
                self.ixx, self.iyy = np.meshgrid(xx, yy)
            elif (xx.ndim == 1) and (yy.ndim == 2):
                self.ixx, tmp = np.meshgrid(xx, yy[:, 0])
                self.iyy = yy
            elif (xx.ndim == 2) and (yy.ndim == 1):
                self.ixx, self.iyy = np.meshgrid(xx, yy)
                tmp, self.iyy = np.meshgrid(xx[0, :], yy)
                self.ixx = xx
            elif (xx.ndim == 2) and (yy.ndim == 2):
                self.ixx = xx
                self.iyy = yy
            else:
                estr  = 'Map: lon (' + vx + '), lat (' + vy + ')'
                estr += ' dimensions not 1D or 2D:'
                print(estr, xx.shape, yy.shape)
                return
            if inv_lon:
                self.ixx = np.fliplr(self.ixx)
            if inv_lat:
                self.iyy = np.flipud(self.iyy)
            self.ivv = vv
            if self.iiglobal:
                # cartopy.contourf needs cyclic longitude for wrap around
                self.ivvc, self.ixxc, self.iyyc = add_cyclic(
                    self.ivv, x=self.ixx, y=self.iyy)
                # # special treatment if fringe points < 1e-4 apart
                # # This works but it did still not display correctly the
                # # test file of GDPS from cuizinart
                # # -> do not do it for the moment
                # if np.ma.allclose(np.ma.sin(np.deg2rad(self.ixxc[:, 0])),
                #                   np.ma.sin(np.deg2rad(self.ixxc[:, -1])),
                #                   atol=1.0e-5):
                #     if np.ma.allclose(self.ixxc[:, 0], self.ixxc[:, -1],
                #                       atol=1.0e-4):
                #         self.ixxc = self.ixxc.astype(float)
                #         self.ixxc[:, -1] = self.ixxc[:, 0] + 360.
            else:
                self.ivvc = self.ivv
                self.ixxc = self.ixx
                self.iyyc = self.iyy
            self.itrans  = ccrs.PlateCarree()
            self.ivmin   = vmin
            self.ivmax   = vmax
            self.icmap   = cmap
            self.ncmap   = mpl.cm.get_cmap(self.icmap).N
            self.ncmap   = self.ncmap if self.ncmap < 256 else 15
            self.iextend = extend
            if mesh:
                try:
                    # vv is matrix notation: (row, col)
                    # self.cc = self.axes.pcolormesh(
                    #     xx, yy, vv, vmin=vmin, vmax=vmax, cmap=cmap,
                    #     shading='nearest')
                    self.cc = self.axes.pcolormesh(
                        self.ixx, self.iyy, self.ivv,
                        vmin=self.ivmin, vmax=self.ivmax,
                        cmap=self.icmap, shading='nearest',
                        transform=self.itrans)
                    # self.cc = self.axes.imshow(
                    #     vv, vmin=vmin, vmax=vmax, cmap=cmap,
                    #     origin='upper', extent=self.img_extent,
                    #     transform=self.itrans)
                    self.cb = self.figure.colorbar(self.cc, fraction=0.05,
                                                   shrink=0.75, pad=0.07,
                                                   extend=self.iextend)
                except Exception:
                    estr  = 'Map pcolormesh: lon (' + vx + '), '
                    estr += ' lat (' + vy + '), var (' + vz + ') shapes do not'
                    estr += ' match:'
                    print(estr, self.ixx.shape, self.iyy.shape, self.ivv.shape)
                    return
            else:
                try:
                    # if 1-D then len(x)==m (columns) and
                    #     len(y)==n (rows): v(n,m)
                    self.cc = self.axes.contourf(
                        self.ixxc, self.iyyc, self.ivvc, self.ncmap,
                        vmin=self.ivmin, vmax=self.ivmax,
                        cmap=self.icmap, extend=self.iextend,
                        transform=self.itrans)
                    self.cb = self.figure.colorbar(self.cc, fraction=0.05,
                                                   shrink=0.75, pad=0.07)
                    # self.cc, = self.axes.plot(yy, vv[0,:])
                except Exception:
                    estr  = 'Map contourf: lon (' + vx + '), lat (' + vy + '),'
                    estr += ' var (' + vz + ') shapes do not match for:'
                    print(estr, self.ixxc.shape, self.iyyc.shape,
                          self.ivvc.shape)
                    return
            self.cb.set_label(vlab)
            self.axes.format_coord = lambda x, y: format_coord_map(
                x, y, self.axes, self.ixx, self.iyy, self.ivv)
        # help(self.figure)
        if self.iiglobal:
            self.axes.set_global()
        if coast:
            self.axes.coastlines()
            self.axes.gridlines(draw_labels=True, linewidth=0,
                                x_inline=False, y_inline=False)
        self.axes.xaxis.set_label_text(xlab)
        self.axes.yaxis.set_label_text(ylab)
        if grid:
            self.axes.gridlines(draw_labels=False,
                                x_inline=False, y_inline=False)
        # redraw
        self.canvas.draw()
        self.toolbar.update()

    # def update(self, frame, isframe=False):
    #     pass

    def update(self, frame, isframe=False):
        """
        Updates data of the current the plot.
        """
        if self.anim_first:
            self.anim.event_source.stop()
            self.anim_running = False
            self.anim_first   = False
            return
        # variable
        v = self.v.get()
        if (v != ''):
            trans_v   = self.trans_v.get()
            mesh      = self.mesh.get()
            rep       = self.repeat.get()
            inv_lon   = self.inv_lon.get()
            inv_lat   = self.inv_lat.get()
            shift_lon = self.shift_lon.get()
            vz = vardim2var(v)
            if vz == self.tname:
                vz = self.tvar
            vv = self.fi.variables[vz]
            # slice
            try:
                it = int(self.vdval[self.iunlim].get())
                if not isframe:
                    if (self.anim_inc == 1) and (it == self.nunlim-1):
                        if rep == 'repeat':
                            it = 0
                        elif rep == 'reflect':
                            self.anim_inc = -1
                            it += self.anim_inc
                        else:  # once
                            self.anim.event_source.stop()
                            self.anim_running = False
                    elif (self.anim_inc == -1) and (it == 0):
                        if rep == 'repeat':
                            it = self.nunlim - 1
                        elif rep == 'reflect':
                            self.anim_inc = 1
                            it += self.anim_inc
                        else:  # once
                            self.anim.event_source.stop()
                            self.anim_running = False
                    else:
                        it += self.anim_inc
            except ValueError:
                it = 0
            self.set_tstep(it)
            vv = get_slice_miss(self, self.vd, vv)
            if vv.ndim < 2:
                self.anim.event_source.stop()
                self.anim_running = False
                return
            if trans_v:
                vv = vv.T
            if shift_lon:
                vv = np.roll(vv, vv.shape[1]//2, axis=1)
            self.ivv = vv
            # set data
            if mesh:
                # update works well on "normal" pcolormesh but not on Cartopy's
                # self.cc.set_array(vv)
                # Both, imshow and pcolormesh need to remove the old
                # image.AxesImage or collections.QuadMesh first and then redraw
                # because the set_data (imshow) and set_array (pcolormesh) do
                # not respect transformations.
                self.cc.remove()
                self.cc = self.axes.pcolormesh(
                    self.ixx, self.iyy, self.ivv,
                    vmin=self.ivmin, vmax=self.ivmax,
                    cmap=self.icmap, shading='nearest',
                    transform=self.itrans)
                # self.cc.remove()
                # self.cc = self.axes.imshow(
                #     vv, vmin=self.ivmin, vmax=self.ivmax, cmap=self.icmap,
                #     origin='upper', extent=self.img_extent,
                #     transform=self.itrans)
            else:
                # http://matplotlib.1069221.n5.nabble.com/update-an-existing-contour-plot-with-new-data-td23889.html
                for coll in self.cc.collections:
                    self.axes.collections.remove(coll)
                if self.iiglobal:
                    # self.ivvc = add_cyclic(self.ivv)
                    self.ivvc, self.ixxc = add_cyclic(
                        self.ivv, x=self.ixx)
                else:
                    self.ivvc = self.ivv
                self.cc = self.axes.contourf(
                    self.ixxc, self.iyyc, self.ivvc, self.ncmap,
                    vmin=self.ivmin, vmax=self.ivmax,
                    cmap=self.icmap, extend=self.iextend,
                    transform=self.itrans)
            self.canvas.draw()
            return self.cc,



#!/usr/bin/env python
"""
Main ncvue window.

This sets up the main notebook window with the plotting panels and
analyses the netcdf file, e.g. determining the unlimited dimensions,
calculating dates, etc.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   ncvMain

History
    * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added check_new_netcdf method that re-initialises all panels if netcdf
      file changed, Jan 2021, Matthias Cuntz

"""
from __future__ import absolute_import, division, print_function
import tkinter as tk
try:
    import tkinter.ttk as ttk
except Exception:
    import sys
    print('Using the themed widget set introduced in Tk 8.5.')
    sys.exit()
import numpy as np
from .ncvutils   import vardim2var
from .ncvscatter import ncvScatter
from .ncvcontour import ncvContour
from .ncvmap     import ncvMap

__all__ = ['ncvMain']


# --------------------------------------------------------------------
# Window with plot panels
#

class ncvMain(ttk.Frame):
    """
    Main ncvue notebook window with the plotting panels.

    Sets up the notebook layout with the panels and analyses the netcdf file,
    e.g. determining the unlimited dimensions, calculating dates, etc. in
    __init__.

    Contains the method to analyse the netcdf file.

    """

    #
    # Window setup
    #

    def __init__(self, master, **kwargs):
        super().__init__(master, **kwargs)
        self.pack(fill=tk.BOTH, expand=1)

        self.name   = 'ncvMain'
        self.master = master      # master window, i.e. root
        self.top    = master.top  # top window

        # Notebook for tabs for future plot types
        self.tabs = ttk.Notebook(self)
        self.tabs.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        self.tab_scatter = ncvScatter(self)
        self.tab_contour = ncvContour(self)
        self.tab_map     = ncvMap(self)

        mapfirst = False
        if self.top.latvar:
            vl = vardim2var(self.top.latvar)
            if np.prod(self.top.fi.variables[vl].shape) > 1:
                mapfirst = True
        if self.top.lonvar:
            vl = vardim2var(self.top.lonvar)
            if np.prod(self.top.fi.variables[vl].shape) > 1:
                mapfirst = True

        if mapfirst:
            self.tabs.add(self.tab_map, text=self.tab_map.name)
        self.tabs.add(self.tab_scatter, text=self.tab_scatter.name)
        self.tabs.add(self.tab_contour, text=self.tab_contour.name)
        if not mapfirst:
            self.tabs.add(self.tab_map, text=self.tab_map.name)

        self.tabs.bind("<<NotebookTabChanged>>", self.check_new_netcdf)
        self.tabs.bind("<Enter>", self.check_new_netcdf)

    #
    # Methods
    #

    def check_new_netcdf(self, event):
        """
        Command called if notebook panel changed or mouse pointer enters a
        window. It checks if netcdf file was changed in any panel of any window
        and re-initialises all plot panels (of the current window).
        """
        if self.tab_scatter.top.fi != self.tab_scatter.fi:
            self.tab_scatter.reinit()
            self.tab_scatter.redraw()
        if self.tab_contour.top.fi != self.tab_contour.fi:
            self.tab_contour.reinit()
            self.tab_contour.redraw()
        if self.tab_map.top.fi != self.tab_map.fi:
            self.tab_map.reinit()
            self.tab_map.redraw()



#!/usr/bin/env python
"""
Contour panel of ncvue.

The panel allows plotting contour or mesh plots of 2D-variables.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   ncvContour

History
    * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Open new netcdf file, communicate via top widget,
      Jan 2021, Matthias Cuntz
    * Write coordinates and value on bottom of plotting canvas,
      May 2021, Matthias Cuntz

"""
from __future__ import absolute_import, division, print_function
import sys
import tkinter as tk
try:
    import tkinter.ttk as ttk
except Exception:
    print('Using the themed widget set introduced in Tk 8.5.')
    sys.exit()
from tkinter import filedialog
import os
import numpy as np
import netCDF4 as nc
from .ncvutils   import clone_ncvmain, format_coord_contour
from .ncvutils   import set_axis_label, vardim2var
from .ncvmethods import analyse_netcdf, get_slice_miss
from .ncvmethods import set_dim_x, set_dim_y, set_dim_z
from .ncvwidgets import add_checkbutton, add_combobox, add_entry, add_imagemenu
from .ncvwidgets import add_spinbox, add_tooltip
import matplotlib
# matplotlib.use('TkAgg')
from matplotlib import pyplot as plt
# plt.style.use('seaborn-darkgrid')
plt.style.use('seaborn-dark')
# plt.style.use('fast')


__all__ = ['ncvContour']


class ncvContour(ttk.Frame):
    """
    Panel for contour plots.

    Sets up the layout with the figure canvas, variable selectors, dimension
    spinboxes, and options in __init__.

    Contains various commands that manage what will be drawn or redrawn if
    something is selected, changed, checked, etc.

    """

    #
    # Panel setup
    #

    def __init__(self, master, **kwargs):
        from functools import partial
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
        from matplotlib.figure import Figure

        super().__init__(master, **kwargs)

        self.name   = 'Contour'
        self.master = master
        self.top    = master.top
        # copy for ease of use
        self.fi     = self.top.fi
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols

        # new window
        self.rowwin = ttk.Frame(self)
        self.rowwin.pack(side=tk.TOP, fill=tk.X)
        self.newfile = ttk.Button(self.rowwin, text="Open File",
                                  command=self.newnetcdf)
        self.newfile.pack(side=tk.LEFT)
        self.newfiletip = add_tooltip(self.newfile, 'Open a new netcdf file')
        self.newwin = ttk.Button(
            self.rowwin, text="New Window",
            command=partial(clone_ncvmain, self.master))
        self.newwin.pack(side=tk.RIGHT)
        self.newwintip = add_tooltip(
            self.newwin, 'Open secondary ncvue window')

        # plotting canvas
        self.figure = Figure(facecolor="white", figsize=(1, 1))
        self.axes   = self.figure.add_subplot(111)
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.draw()
        # pack
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        # grid instead of pack - does not work
        # self.canvas.get_tk_widget().grid(column=0, row=0,
        #     sticky=(tk.N, tk.S, tk.E, tk.W))
        # self.canvas.get_tk_widget().columnconfigure(0, weight=1)
        # self.canvas.get_tk_widget().rowconfigure(0, weight=1)

        # matplotlib toolbar
        self.toolbar = NavigationToolbar2Tk(self.canvas, self)
        self.toolbar.update()
        self.toolbar.pack(side=tk.TOP, fill=tk.X)

        # selections and options
        columns = [''] + self.cols

        allcmaps = plt.colormaps()
        self.cmaps  = [ i for i in allcmaps if not i.endswith('_r') ]
        self.cmaps.sort()
        # self.imaps  = [ tk.PhotoImage(file=os.path.dirname(__file__) +
        #                               '/images/' + i + '.png')
        #                 for i in self.cmaps ]
        bundle_dir = getattr(sys, '_MEIPASS',
                             os.path.abspath(os.path.dirname(__file__)))
        self.imaps  = [ tk.PhotoImage(file=bundle_dir +
                                      '/images/' + i + '.png')
                        for i in self.cmaps ]

        # 1. row
        # z-axis selection
        self.rowzz = ttk.Frame(self)
        self.rowzz.pack(side=tk.TOP, fill=tk.X)
        self.blockz = ttk.Frame(self.rowzz)
        self.blockz.pack(side=tk.LEFT)
        self.rowz = ttk.Frame(self.blockz)
        self.rowz.pack(side=tk.TOP, fill=tk.X)
        self.zlbl = tk.StringVar()
        self.zlbl.set("z")
        zlab = ttk.Label(self.rowz, textvariable=self.zlbl)
        zlab.pack(side=tk.LEFT)
        self.bprev_z = ttk.Button(self.rowz, text="<", width=1,
                                  command=self.prev_z)
        self.bprev_z.pack(side=tk.LEFT)
        self.bprev_ztip = add_tooltip(self.bprev_z, 'Previous variable')
        self.bnext_z = ttk.Button(self.rowz, text=">", width=1,
                                  command=self.next_z)
        self.bnext_z.pack(side=tk.LEFT)
        self.bnext_ztip = add_tooltip(self.bnext_z, 'Next variable')
        self.z = ttk.Combobox(self.rowz, values=columns, width=25)
        self.z.bind("<<ComboboxSelected>>", self.selected_z)
        self.z.pack(side=tk.LEFT)
        self.ztip = add_tooltip(self.z, 'Choose variable')
        self.trans_zlbl, self.trans_z, self.trans_ztip = add_checkbutton(
            self.rowz, label="transpose z", value=False, command=self.checked,
            tooltip="Transpose matrix")
        spacez = ttk.Label(self.rowz, text=" "*1)
        spacez.pack(side=tk.LEFT)
        self.zminlbl, self.zmin, self.zmintip = add_entry(
            self.rowz, label="zmin", text='None', width=7,
            command=self.entered_z,
            tooltip="Minimal display value. Free scaling if 'None'.")
        self.zmaxlbl, self.zmax, self.zmaxtip = add_entry(
            self.rowz, label="zmax", text='None', width=7,
            command=self.entered_z,
            tooltip="Maximal display value. Free scaling if 'None'.")
        # levels z
        self.rowzd = ttk.Frame(self.blockz)
        self.rowzd.pack(side=tk.TOP, fill=tk.X)
        self.zdlblval = []
        self.zdlbl    = []
        self.zdval    = []
        self.zd       = []
        self.zdtip    = []
        for i in range(self.maxdim):
            zdlblval, zdlbl, zdval, zd, zdtip = add_spinbox(
                self.rowzd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_z, state=tk.DISABLED, tooltip="None")
            self.zdlblval.append(zdlblval)
            self.zdlbl.append(zdlbl)
            self.zdval.append(zdval)
            self.zd.append(zd)
            self.zdtip.append(zdtip)

        # 2. row
        # x-axis selection
        self.rowxy = ttk.Frame(self)
        self.rowxy.pack(side=tk.TOP, fill=tk.X)
        self.blockx = ttk.Frame(self.rowxy)
        self.blockx.pack(side=tk.LEFT)
        self.rowx = ttk.Frame(self.blockx)
        self.rowx.pack(side=tk.TOP, fill=tk.X)
        self.xlbl, self.x, self.xtip = add_combobox(
            self.rowx, label="x", values=columns, command=self.selected_x,
            tooltip="Choose variable of x-axis.\nTake index if 'None' (fast).")
        self.inv_xlbl, self.inv_x, self.inv_xtip = add_checkbutton(
            self.rowx, label="invert x", value=False, command=self.checked,
            tooltip="Invert x-axis")
        self.rowxd = ttk.Frame(self.blockx)
        self.rowxd.pack(side=tk.TOP, fill=tk.X)
        self.xdlblval = []
        self.xdlbl    = []
        self.xdval    = []
        self.xd       = []
        self.xdtip    = []
        for i in range(self.maxdim):
            xdlblval, xdlbl, xdval, xd, xdtip = add_spinbox(
                self.rowxd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_x, state=tk.DISABLED, tooltip="None")
            self.xdlblval.append(xdlblval)
            self.xdlbl.append(xdlbl)
            self.xdval.append(xdval)
            self.xd.append(xd)
            self.xdtip.append(xdtip)
        # y-axis selection
        spacex = ttk.Label(self.rowxy, text=" "*3)
        spacex.pack(side=tk.LEFT)
        self.blocky = ttk.Frame(self.rowxy)
        self.blocky.pack(side=tk.LEFT)
        self.rowy = ttk.Frame(self.blocky)
        self.rowy.pack(side=tk.TOP, fill=tk.X)
        self.ylbl, self.y, self.ytip = add_combobox(
            self.rowy, label="y", values=columns, command=self.selected_y,
            tooltip="Choose variable of y-axis.\nTake index if 'None'.")
        self.inv_ylbl, self.inv_y, self.inv_ytip = add_checkbutton(
            self.rowy, label="invert y", value=False, command=self.checked,
            tooltip="Invert y-axis")
        self.rowyd = ttk.Frame(self.blocky)
        self.rowyd.pack(side=tk.TOP, fill=tk.X)
        self.ydlblval = []
        self.ydlbl    = []
        self.ydval    = []
        self.yd       = []
        self.ydtip    = []
        for i in range(self.maxdim):
            ydlblval, ydlbl, ydval, yd, ydtip = add_spinbox(
                self.rowyd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_y, state=tk.DISABLED, tooltip="None")
            self.ydlblval.append(ydlblval)
            self.ydlbl.append(ydlbl)
            self.ydval.append(ydval)
            self.yd.append(yd)
            self.ydtip.append(ydtip)

        # 3. row
        # options
        self.rowcmap = ttk.Frame(self)
        self.rowcmap.pack(side=tk.TOP, fill=tk.X)
        self.cmaplbl, self.cmap, self.cmaptip = add_imagemenu(
            self.rowcmap, label="cmap", values=self.cmaps,
            images=self.imaps, command=self.selected_cmap,
            tooltip="Choose colormap")
        self.cmap['text']  = 'RdYlBu'
        self.cmap['image'] = self.imaps[self.cmaps.index('RdYlBu')]
        self.rev_cmaplbl, self.rev_cmap, self.rev_cmaptip = add_checkbutton(
            self.rowcmap, label="reverse cmap", value=False,
            command=self.checked,
            tooltip="Reverse colormap")
        self.meshlbl, self.mesh, self.meshtip = add_checkbutton(
            self.rowcmap, label="mesh", value=True,
            command=self.checked,
            tooltip="Pseudocolor plot if checked, plot contours if unchecked")
        self.gridlbl, self.grid, self.gridtip = add_checkbutton(
            self.rowcmap, label="grid", value=False,
            command=self.checked,
            tooltip="Draw major grid lines")

    #
    # Bindings
    #

    def checked(self):
        """
        Command called if any checkbutton was checked or unchecked.

        Redraws plot.
        """
        self.redraw()

    def entered_z(self, event):
        """
        Command called if values for `zmin`/`zmax` were entered.

        Triggering `event` was bound to entry.

        Redraws plot.
        """
        self.redraw()

    def next_z(self):
        """
        Command called if next button for the plotting variable was pressed.

        Resets `zmin`/`zmax` and z-dimensions, resets `x` and `y` variables
        as well as their options and dimensions. Redraws plot.
        """
        z = self.z.get()
        cols = self.z["values"]
        idx  = cols.index(z)
        idx += 1
        if idx < len(cols):
            self.z.set(cols[idx])
            self.zmin.set('None')
            self.zmax.set('None')
            set_dim_z(self)
            self.x.set('')
            self.y.set('')
            self.inv_x.set(0)
            self.inv_y.set(0)
            set_dim_x(self)
            set_dim_y(self)
            self.redraw()

    def prev_z(self):
        """
        Command called if previous button for the plotting variable was
        pressed.

        Resets `zmin`/`zmax` and z-dimensions, resets `x` and `y` variables
        as well as their options and dimensions. Redraws plot.
        """
        z = self.z.get()
        cols = self.z["values"]
        idx  = cols.index(z)
        idx -= 1
        if idx > 0:
            self.z.set(cols[idx])
            self.zmin.set('None')
            self.zmax.set('None')
            set_dim_z(self)
            self.x.set('')
            self.y.set('')
            self.inv_x.set(0)
            self.inv_y.set(0)
            set_dim_x(self)
            set_dim_y(self)
            self.redraw()

    def newnetcdf(self):
        """
        Open a new netcdf file and connect it to top.
        """
        # get new netcdf file name
        ncfile = filedialog.askopenfilename(
            parent=self, title='Choose netcdf file', multiple=False)
        if ncfile:
            # close old netcdf file
            if self.top.fi:
                self.top.fi.close()
            # reset empty defaults of top
            self.top.dunlim = ''      # name of unlimited dimension
            self.top.time   = None    # datetime variable
            self.top.tname  = ''      # datetime variable name
            self.top.tvar   = ''      # datetime variable name in netcdf
            self.top.dtime  = None    # decimal year
            self.top.latvar = ''      # name of latitude variable
            self.top.lonvar = ''      # name of longitude variable
            self.top.latdim = ''      # name of latitude dimension
            self.top.londim = ''      # name of longitude dimension
            self.top.maxdim = 0       # maximum num of dims of all variables
            self.top.cols   = []      # variable list
            # open new netcdf file
            self.top.fi = nc.Dataset(ncfile, 'r')
            analyse_netcdf(self.top)
            # reset panel
            self.reinit()
            self.redraw()

    def selected_cmap(self, value):
        """
        Command called if cmap was chosen from menu.

        `value` is the chosen colormap.

        Sets text and image on the menubutton.
        """
        self.cmap['text']  = value
        self.cmap['image'] = self.imaps[self.cmaps.index(value)]
        self.redraw()

    def selected_x(self, event):
        """
        Command called if x-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `x` options and dimensions. Redraws plot.
        """
        self.inv_x.set(0)
        set_dim_x(self)
        self.redraw()

    def selected_y(self, event):
        """
        Command called if y-variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `y` options and dimensions. Redraws plot.
        """
        self.inv_y.set(0)
        set_dim_y(self)
        self.redraw()

    def selected_z(self, event):
        """
        Command called if plotting variable was selected with combobox.

        Triggering `event` was bound to the combobox.

        Resets `zmin`/`zmax` and z-dimensions, resets `x` and `y` variables
        as well as their options and dimensions. Redraws plot.
        """
        self.x.set('')
        self.y.set('')
        self.inv_x.set(0)
        self.inv_y.set(0)
        self.zmin.set('None')
        self.zmax.set('None')
        set_dim_x(self)
        set_dim_y(self)
        set_dim_z(self)
        self.redraw()

    def spinned_x(self, event=None):
        """
        Command called if spinbox of x-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        self.redraw()

    def spinned_y(self, event=None):
        """
        Command called if spinbox of y-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        self.redraw()

    def spinned_z(self, event=None):
        """
        Command called if spinbox of z-dimensions was changed.

        Triggering `event` was bound to the spinbox.

        Redraws plot.
        """
        self.redraw()

    #
    # Methods
    #

    def reinit(self):
        """
        Reinitialise the panel from top.
        """
        # reinit from top
        self.fi     = self.top.fi
        self.miss   = self.top.miss
        self.dunlim = self.top.dunlim
        self.time   = self.top.time
        self.tname  = self.top.tname
        self.tvar   = self.top.tvar
        self.dtime  = self.top.dtime
        self.latvar = self.top.latvar
        self.lonvar = self.top.lonvar
        self.latdim = self.top.latdim
        self.londim = self.top.londim
        self.maxdim = self.top.maxdim
        self.cols   = self.top.cols
        # reset dimensions
        for ll in self.zdlbl:
            ll.destroy()
        for ll in self.zd:
            ll.destroy()
        self.zdlblval = []
        self.zdlbl    = []
        self.zdval    = []
        self.zd       = []
        self.zdtip    = []
        for i in range(self.maxdim):
            zdlblval, zdlbl, zdval, zd, zdtip = add_spinbox(
                self.rowzd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_z, state=tk.DISABLED, tooltip="None")
            self.zdlblval.append(zdlblval)
            self.zdlbl.append(zdlbl)
            self.zdval.append(zdval)
            self.zd.append(zd)
            self.zdtip.append(zdtip)
        for ll in self.xdlbl:
            ll.destroy()
        for ll in self.xd:
            ll.destroy()
        self.xdlblval = []
        self.xdlbl    = []
        self.xdval    = []
        self.xd       = []
        self.xdtip    = []
        for i in range(self.maxdim):
            xdlblval, xdlbl, xdval, xd, xdtip = add_spinbox(
                self.rowxd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_x, state=tk.DISABLED, tooltip="None")
            self.xdlblval.append(xdlblval)
            self.xdlbl.append(xdlbl)
            self.xdval.append(xdval)
            self.xd.append(xd)
            self.xdtip.append(xdtip)
        for ll in self.ydlbl:
            ll.destroy()
        for ll in self.yd:
            ll.destroy()
        self.ydlblval = []
        self.ydlbl    = []
        self.ydval    = []
        self.yd       = []
        self.ydtip    = []
        for i in range(self.maxdim):
            ydlblval, ydlbl, ydval, yd, ydtip = add_spinbox(
                self.rowyd, label=str(i), values=(0,), wrap=True,
                command=self.spinned_y, state=tk.DISABLED, tooltip="None")
            self.ydlblval.append(ydlblval)
            self.ydlbl.append(ydlbl)
            self.ydval.append(ydval)
            self.yd.append(yd)
            self.ydtip.append(ydtip)
        # set variables
        columns = [''] + self.cols
        self.z['values'] = columns
        self.z.set(columns[0])
        self.zmin.set('None')
        self.zmax.set('None')
        self.x['values'] = columns
        self.x.set(columns[0])
        self.y['values'] = columns
        self.y.set(columns[0])

    #
    # Plotting
    #

    def redraw(self):
        """
        Redraws the plot.

        Reads `x`, `y`, `z` variable names, the current settings of
        their dimension spinboxes, as well as all other plotting options.
        Then redraws the plot.
        """
        # get all states
        # rowz
        z = self.z.get()
        trans_z = self.trans_z.get()
        zmin = self.zmin.get()
        if zmin == 'None':
            zmin = None
        else:
            zmin = float(zmin)
        zmax = self.zmax.get()
        if zmax == 'None':
            zmax = None
        else:
            zmax = float(zmax)
        # rowxy
        x = self.x.get()
        y = self.y.get()
        inv_x = self.inv_x.get()
        inv_y = self.inv_y.get()
        # rowcmap
        cmap = self.cmap['text']
        rev_cmap = self.rev_cmap.get()
        mesh = self.mesh.get()
        grid = self.grid.get()
        # Clear figure instead of axes because colorbar is on figure
        # Have to add axes again.
        self.figure.clear()
        self.axes = self.figure.add_subplot(111)
        xlim  = [None, None]
        ylim  = [None, None]
        # set x, y, axes labels
        vx = 'None'
        vy = 'None'
        vz = 'None'
        if (z != ''):
            # z axis
            vz = vardim2var(z)
            if vz == self.tname:
                # should throw an error later
                if mesh:
                    zz = self.dtime
                    zlab = 'Year'
                else:
                    zz = self.time
                    zlab = 'Date'
            else:
                zz = self.fi.variables[vz]
                zlab = set_axis_label(zz)
            zz = get_slice_miss(self, self.zd, zz)
            # both contourf and pcolormesh assume (row,col),
            # so transpose by default
            if not trans_z:
                zz = zz.T
        if (y != ''):
            # y axis
            vy = vardim2var(y)
            if vy == self.tname:
                if mesh:
                    yy = self.dtime
                    ylab = 'Year'
                else:
                    yy = self.time
                    ylab = 'Date'
            else:
                yy   = self.fi.variables[vy]
                ylab = set_axis_label(yy)
            yy = get_slice_miss(self, self.yd, yy)
        if (x != ''):
            # x axis
            vx = vardim2var(x)
            if vx == self.tname:
                if mesh:
                    xx = self.dtime
                    xlab = 'Year'
                else:
                    xx = self.time
                    xlab = 'Date'
            else:
                xx   = self.fi.variables[vx]
                xlab = set_axis_label(xx)
            xx = get_slice_miss(self, self.xd, xx)
        # set z to nan if not selected
        if (z == ''):
            if (x != ''):
                nx = xx.shape[0]
            else:
                nx = 1
            if (y != ''):
                ny = yy.shape[0]
            else:
                ny = 1
            zz = np.ones((ny, nx)) * np.nan
            zlab = ''
        if zz.ndim < 2:
            estr  = 'Contour: z (' + vz + ') is not 2-dimensional:'
            print(estr, zz.shape)
            return
        # set x and y to index if not selected
        if (x == ''):
            nx = zz.shape[1]
            xx = np.arange(nx)
            xlab = ''
        if (y == ''):
            ny = zz.shape[0]
            yy = np.arange(ny)
            ylab = ''
        # plot options
        if rev_cmap:
            cmap = cmap + '_r'
        # plot
        # cc = self.axes.imshow(zz[:, ::-1], aspect='auto', cmap=cmap,
        #                       interpolation='none')
        # cc = self.axes.matshow(zz[:, ::-1], aspect='auto', cmap=cmap,
        #                        interpolation='none')
        extend = 'neither'
        if zmin is not None:
            zz = np.maximum(zz, zmin)
            if zmax is None:
                extend = 'min'
            else:
                extend = 'both'
        if zmax is not None:
            zz = np.minimum(zz, zmax)
            if zmin is None:
                extend = 'max'
            else:
                extend = 'both'
        if mesh:
            try:
                # zz is matrix notation: (row, col)
                cc = self.axes.pcolormesh(xx, yy, zz, vmin=zmin, vmax=zmax,
                                          cmap=cmap, shading='nearest')
                cb = self.figure.colorbar(cc, fraction=0.05, shrink=0.75,
                                          extend=extend)
            except Exception:
                estr  = 'Contour: x (' + vx + '), y (' + vy + '),'
                estr += ' z (' + vz + ') shapes do not match for'
                estr += ' pcolormesh:'
                print(estr, xx.shape, yy.shape, zz.shape)
                return
        else:
            try:
                # if 1-D then len(x)==m (columns) and len(y)==n (rows): z(n,m)
                cc = self.axes.contourf(xx, yy, zz, vmin=zmin, vmax=zmax,
                                        cmap=cmap, extend=extend)
                cb = self.figure.colorbar(cc, fraction=0.05, shrink=0.75)
            except Exception:
                estr  = 'Contour: x (' + vx + '), y (' + vy + '),'
                estr += ' z (' + vz + ') shapes do not match for'
                estr += ' contourf:'
                print(estr, xx.shape, yy.shape, zz.shape)
                return
        # help(self.figure)
        cb.set_label(zlab)
        self.axes.xaxis.set_label_text(xlab)
        self.axes.yaxis.set_label_text(ylab)
        self.axes.format_coord = lambda x, y: format_coord_contour(
            x, y, self.axes, xx, yy, zz)
        # # Does not work
        # # might do it by hand, i.e. get ticks and use axhline and axvline
        # self.axes.grid(True, lw=5, color='k', zorder=100)
        # self.axes.set_zorder(100)
        # self.axes.xaxis.grid(True, zorder=999)
        # self.axes.yaxis.grid(True, zorder=999)
        xlim = self.axes.get_xlim()
        ylim = self.axes.get_ylim()
        # invert axes
        if inv_x:
            if (xlim[0] is not None):
                xlim = xlim[::-1]
                self.axes.set_xlim(xlim)
        if inv_y:
            if (ylim[0] is not None):
                ylim = ylim[::-1]
                self.axes.set_ylim(ylim)
        # draw grid lines
        self.axes.grid(False)
        xticks = np.array(self.axes.get_xticks())
        yticks = np.array(self.axes.get_yticks())
        if grid:
            ii = np.where((xticks > min(xlim)) & (xticks < max(xlim)))[0]
            if ii.size > 0:
                ggx = self.axes.vlines(xticks[ii], ylim[0], ylim[1],
                                       colors='w', linestyles='solid',
                                       linewidth=0.5)
            ii = np.where((yticks > min(ylim)) & (yticks < max(ylim)))[0]
            if ii.size > 0:
                ggy = self.axes.hlines(yticks[ii], xlim[0], xlim[1],
                                       colors='w', linestyles='solid',
                                       linewidth=0.5)
        # redraw
        self.canvas.draw()
        self.toolbar.update()



#!/usr/bin/env python
"""
Common methods for panels of ncvue.

Methods normally belong to a class and be called like self.method(args).
We have several panels in ncvue that would have the same methods such as
getting slices from arrays. These are hence gathered here, all start with
the first argument self, which is the class instance, and are called like
method(self, args).

The methods could also be put onto a common class (based on ttk.Frame) on
which ncvContour, etc. would then be based.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following methods are provided:

.. autosummary::
   analyse_netcdf
   get_miss
   get_slice_miss
   set_dim_lat
   set_dim_lon
   set_dim_var
   set_dim_x
   set_dim_y
   set_dim_y2
   set_dim_z

History
    * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Slice arrays with slice function rather than numpy.take,
      Dec 2020, Matthias Cuntz
    * Moved individual get_slice? methods for x, y, y2, z as general get_slice
      function to ncvutils, Dec 2020, Matthias Cuntz
    * Added convenience method get_slice_miss, Dec 2020, Matthias Cuntz
    * set_dim_lon, set_dim_lat, set_dim_var for Map panel,
      Jan 2021, Matthias Cuntz
    * Set latdim, londim to all on map var, determined in ncvMain,
      Jan 2021, Matthias Cuntz
    * Catch non numpy.dtype in set_miss, Jan 2021, Matthias Cuntz
    * Catch variables that have only one string or similar,
      Jan 2021, Matthias Cuntz
    * Added tooltip to dimensions, Jan 2021, Matthias Cuntz
    * Added analyse_netcdf from ncvmain, Jan 2021, Matthias Cuntz
    * Use dlblval instead of dlbl to set dimension labels,
      Jan 2021, Matthias Cuntz
    * Make self.time numpy's datetime64[ms] format, May 2021, Matthias Cuntz
    * Add numpy's datetime64[ms] to missing values, May 2021, Matthias Cuntz
    * Removed bug in detection of lon/lat (analyse_netcdf),
      Oct 2021, Matthias Cuntz
    * Identify lon/lat also by axis attributes x/y or X/Y (analyse_netcdf),
      Oct 2021, Matthias Cuntz
    * Do not default the unlimited dimension to 'all' if no lon/lat were found,
      (get_dim_var) Oct 2021, Matthias Cuntz

"""
from __future__ import absolute_import, division, print_function
import tkinter as tk
import numpy as np
from .ncvutils import DIMMETHODS, get_slice, set_miss, spinbox_values
from .ncvutils import vardim2var, zip_dim_name_length
import netCDF4 as nc
# nc.default_fillvals but with keys as variables['var'].dtype
nctypes = [ np.dtype(i) for i in nc.default_fillvals ]
ncfill  = dict(zip(nctypes, list(nc.default_fillvals.values())))
ncfill.update({np.dtype('O'): np.nan})
ncfill.update({np.dtype('<M8[ms]'): np.datetime64('NaT')})


__all__ = ['analyse_netcdf', 'analyse_netcdf_xarray',
           'get_miss', 'get_slice_miss',
           'set_dim_lat', 'set_dim_lon', 'set_dim_var',
           'set_dim_x', 'set_dim_y', 'set_dim_y2', 'set_dim_z']


#
# Analyse netcdf file
#

def analyse_netcdf(self):
    """
    Analyse a netcdf file for the unlimited dimension, calculating datetime,
    variables, latitudes/longitudes variables and dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    Set variables:
        self.dunlim,
        self.time, self.tname, self.tvar, self.dtime,
        self.cols,
        self.latvar, self.lonvar, self.latdim, self.londim

    Examples
    --------
    >>> analyse_netcdf(self)

    """
    import datetime as dt
    try:
        import cftime as cf
    except ModuleNotFoundError:
        import netCDF4 as cf
    #
    # search unlimited dimension
    self.dunlim = ''
    for dd in self.fi.dimensions:
        if self.fi.dimensions[dd].isunlimited():
            self.dunlim = dd
            break
    #
    # search for time variable and make datetime variable
    self.time  = None
    self.tname = ''
    self.tvar  = ''
    self.dtime = None
    for vv in self.fi.variables:
        isunlim = False
        if self.dunlim:
            if vv.lower() == self.fi.dimensions[self.dunlim].name.lower():
                isunlim = True
        if ( isunlim or vv.lower().startswith('time_') or
             (vv.lower() == 'time') or (vv.lower() == 'datetime') or
             (vv.lower() == 'date') ):
            self.tvar = vv
            if vv.lower() == 'datetime':
                self.tname = 'date'
            else:
                self.tname = 'datetime'
            try:
                tunit = self.fi.variables[self.tvar].units
            except AttributeError:
                tunit = ''
            # assure 01, etc. if values < 10
            if tunit.find('since') > 0:
                tt = tunit.split()
                dd = tt[2].split('-')
                dd[0] = ('000'+dd[0])[-4:]
                dd[1] = ('0'+dd[1])[-2:]
                dd[2] = ('0'+dd[1])[-2:]
                tt[2] = '-'.join(dd)
                tunit = ' '.join(tt)
            try:
                tcal = self.fi.variables[self.tvar].calendar
            except AttributeError:
                tcal = 'standard'
            time = self.fi.variables[self.tvar][:]
            # time dimension "day as %Y%m%d.%f" from cdo.
            if ' as ' in tunit:
                itunit = tunit.split()[2]
                dtime = []
                for tt in time:
                    stt = str(tt).split('.')
                    sstt = ('00'+stt[0])[-8:] + '.' + stt[1]
                    dtime.append(dt.datetime.strptime(sstt, itunit))
                ntime = cf.date2num(dtime,
                                    'days since 0001-01-01 00:00:00')
                self.dtime = cf.num2date(ntime,
                                         'days since 0001-01-01 00:00:00')
            else:
                try:
                    self.dtime = cf.num2date(time, tunit, calendar=tcal)
                except ValueError:
                    self.dtime = None
            if self.dtime is not None:
                ntime = len(self.dtime)
                if (tcal == '360_day'):
                    ndays = [360.]*ntime
                elif (tcal == '365_day'):
                    ndays = [365.]*ntime
                elif (tcal == 'noleap'):
                    ndays = [365.]*ntime
                elif (tcal == '366_day'):
                    ndays = [366.]*ntime
                elif (tcal == 'all_leap'):
                    ndays = [366.]*ntime
                else:
                    ndays = [ 365. +
                              float((((t.year%4) == 0) &
                                     ((t.year%100) != 0)) |
                                    ((t.year%400) == 0))
                              for t in self.dtime ]
                self.dtime = np.array([
                    t.year +
                    (t.dayofyr-1 + t.hour / 24. +
                     t.minute / 1440 + t.second / 86400.) / ndays[i]
                    for i, t in enumerate(self.dtime) ])
            # make datetime variable
            if self.time is None:
                try:
                    ttime = cf.num2date(
                        time, tunit, calendar=tcal,
                        only_use_cftime_datetimes=False,
                        only_use_python_datetimes=True)
                    self.time = np.array([ dd.isoformat()
                                           for dd in ttime ],
                                         dtype='datetime64[ms]')
                except:
                    self.time = None
            if self.time is None:
                try:
                    # self.time = cf.num2date(time, tunit,
                    ttime = cf.num2date(time, tunit,
                                        calendar=tcal)
                    self.time = np.array([ dd.isoformat()
                                           for dd in ttime ],
                                         dtype='datetime64[ms]')
                except:
                    self.time = None
            if self.time is None:
                # if not possible use decimal year
                self.time = self.dtime
            if self.time is None:
                # could not interpret time at all,
                # e.g. if units = "months since ..."
                self.time  = time
                self.dtime = time
            # print('time:  ', self.time)
            # print('dtime: ', self.dtime)
            break
    #
    # construct list of variable names with dimensions
    if self.time is not None:
        addt = [
            self.tname + ' ' +
            str(tuple(zip_dim_name_length(self.fi.variables[self.tvar])))]
        self.cols += addt
    ivars = []
    for vv in self.fi.variables:
        # ss = self.fi.variables[vv].shape
        ss = tuple(zip_dim_name_length(self.fi.variables[vv]))
        self.maxdim = max(self.maxdim, len(ss))
        ivars.append((vv, ss, len(ss)))
    self.cols += sorted([ vv[0] + ' ' + str(vv[1])
                          for vv in ivars ])
    #
    # search for lat/lon variables
    self.latvar = ''
    self.lonvar = ''
    # first sweep: *name must be "latitude" and
    #              units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'latitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'longitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # second sweep: name must start with lat and
    #               units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lat':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lon':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # third sweep: name must contain lat and
    #              units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lat') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lon') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # fourth sweep: axis is 'Y' or 'y'
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                saxis = self.fi.variables[vv].axis
            except AttributeError:
                saxis = ''
            if saxis.lower() == 'y':
                self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                saxis = self.fi.variables[vv].axis
            except AttributeError:
                saxis = ''
            if saxis.lower() == 'x':
                self.lonvar = vv
    # fifth sweep: same as first but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'latitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'longitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    # sixth sweep: same as second but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lat':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lon':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    # seventh sweep: same as third but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lat') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lon') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    #
    # determine lat/lon dimensions
    self.latdim = ''
    self.londim = ''
    if self.latvar:
        latshape = self.fi.variables[self.latvar].shape
        if (len(latshape) < 1) or (len(latshape) > 2):
            estr  = 'Something went wrong determining lat/lon:'
            estr += ' latitude variable is not 1D or 2D.'
            print(estr)
            estr = 'latitude variable with dimensions:'
            ldim = self.fi.variables[self.latvar].dimensions
            print(estr, self.latvar, ldim)
            self.latvar = ''
        else:
            self.latdim = self.fi.variables[self.latvar].dimensions[0]
    if self.lonvar:
        lonshape = self.fi.variables[self.lonvar].shape
        if len(lonshape) == 1:
            self.londim = self.fi.variables[self.lonvar].dimensions[0]
        elif len(lonshape) == 2:
            self.londim = self.fi.variables[self.lonvar].dimensions[1]
        else:
            estr  = 'Something went wrong determining lat/lon:'
            estr += ' longitude variable is not 1D or 2D.'
            print(estr)
            estr = 'longitude variable with dimensions:'
            ldim = self.fi.variables[self.lonvar].dimensions
            print(estr, self.lonvar, ldim)
            self.lonvar = ''
    #
    # add units to lat/lon name
    if self.latvar:
        idim = tuple(zip_dim_name_length(self.fi.variables[self.latvar]))
        self.latvar = self.latvar + ' ' + str(idim)
    if self.lonvar:
        idim = tuple(zip_dim_name_length(self.fi.variables[self.lonvar]))
        self.lonvar = self.lonvar + ' ' + str(idim)


#
# Analyse netcdf file with xarray
#

def analyse_netcdf_xarray(self):
    """
    Analyse netcdf files for the unlimited dimension, calculating datetime,
    variables, latitudes/longitudes variables and dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    Set variables:
        self.dunlim,
        self.time, self.tname, self.tvar, self.dtime,
        self.cols,
        self.latvar, self.lonvar, self.latdim, self.londim

    Examples
    --------
    >>> analyse_netcdf(self)

    """
    import datetime as dt
    try:
        import cftime as cf
    except ModuleNotFoundError:
        import netCDF4 as cf
    #
    # search unlimited dimension
    self.dunlim = ''
    for dd in self.fi.dimensions:
        if self.fi.dimensions[dd].isunlimited():
            self.dunlim = dd
            break
    #
    # search for time variable and make datetime variable
    self.time  = None
    self.tname = ''
    self.tvar  = ''
    self.dtime = None
    for vv in self.fi.variables:
        isunlim = False
        if self.dunlim:
            if vv.lower() == self.fi.dimensions[self.dunlim].name.lower():
                isunlim = True
        if ( isunlim or vv.lower().startswith('time_') or
             (vv.lower() == 'time') or (vv.lower() == 'datetime') or
             (vv.lower() == 'date') ):
            self.tvar = vv
            if vv.lower() == 'datetime':
                self.tname = 'date'
            else:
                self.tname = 'datetime'
            try:
                tunit = self.fi.variables[self.tvar].units
            except AttributeError:
                tunit = ''
            # assure 01, etc. if values < 10
            if tunit.find('since') > 0:
                tt = tunit.split()
                dd = tt[2].split('-')
                dd[0] = ('000'+dd[0])[-4:]
                dd[1] = ('0'+dd[1])[-2:]
                dd[2] = ('0'+dd[1])[-2:]
                tt[2] = '-'.join(dd)
                tunit = ' '.join(tt)
            try:
                tcal = self.fi.variables[self.tvar].calendar
            except AttributeError:
                tcal = 'standard'
            time = self.fi.variables[self.tvar][:]
            # time dimension "day as %Y%m%d.%f" from cdo.
            if ' as ' in tunit:
                itunit = tunit.split()[2]
                dtime = []
                for tt in time:
                    stt = str(tt).split('.')
                    sstt = ('00'+stt[0])[-8:] + '.' + stt[1]
                    dtime.append(dt.datetime.strptime(sstt, itunit))
                ntime = cf.date2num(dtime,
                                    'days since 0001-01-01 00:00:00')
                self.dtime = cf.num2date(ntime,
                                         'days since 0001-01-01 00:00:00')
            else:
                try:
                    self.dtime = cf.num2date(time, tunit, calendar=tcal)
                except ValueError:
                    self.dtime = None
            if self.dtime is not None:
                ntime = len(self.dtime)
                if (tcal == '360_day'):
                    ndays = [360.]*ntime
                elif (tcal == '365_day'):
                    ndays = [365.]*ntime
                elif (tcal == 'noleap'):
                    ndays = [365.]*ntime
                elif (tcal == '366_day'):
                    ndays = [366.]*ntime
                elif (tcal == 'all_leap'):
                    ndays = [366.]*ntime
                else:
                    ndays = [ 365. +
                              float((((t.year%4) == 0) &
                                     ((t.year%100) != 0)) |
                                    ((t.year%400) == 0))
                              for t in self.dtime ]
                self.dtime = np.array([
                    t.year +
                    (t.dayofyr-1 + t.hour / 24. +
                     t.minute / 1440 + t.second / 86400.) / ndays[i]
                    for i, t in enumerate(self.dtime) ])
            # make datetime variable
            if self.time is None:
                try:
                    ttime = cf.num2date(
                        time, tunit, calendar=tcal,
                        only_use_cftime_datetimes=False,
                        only_use_python_datetimes=True)
                    self.time = np.array([ dd.isoformat()
                                           for dd in ttime ],
                                         dtype='datetime64[ms]')
                except:
                    self.time = None
            if self.time is None:
                try:
                    # self.time = cf.num2date(time, tunit,
                    ttime = cf.num2date(time, tunit,
                                        calendar=tcal)
                    self.time = np.array([ dd.isoformat()
                                           for dd in ttime ],
                                         dtype='datetime64[ms]')
                except:
                    self.time = None
            if self.time is None:
                # if not possible use decimal year
                self.time = self.dtime
            if self.time is None:
                # could not interpret time at all,
                # e.g. if units = "months since ..."
                self.time  = time
                self.dtime = time
            # print('time:  ', self.time)
            # print('dtime: ', self.dtime)
            break
    #
    # construct list of variable names with dimensions
    if self.time is not None:
        addt = [
            self.tname + ' ' +
            str(tuple(zip_dim_name_length(self.fi.variables[self.tvar])))]
        self.cols += addt
    ivars = []
    for vv in self.fi.variables:
        # ss = self.fi.variables[vv].shape
        ss = tuple(zip_dim_name_length(self.fi.variables[vv]))
        self.maxdim = max(self.maxdim, len(ss))
        ivars.append((vv, ss, len(ss)))
    self.cols += sorted([ vv[0] + ' ' + str(vv[1])
                          for vv in ivars ])
    #
    # search for lat/lon variables
    self.latvar = ''
    self.lonvar = ''
    # first sweep: *name must be "latitude" and
    #              units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'latitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'longitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # second sweep: name must start with lat and
    #               units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lat':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lon':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # third sweep: name must contain lat and
    #              units must be "degrees_north"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lat') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_north':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lon') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees_east':
                    self.lonvar = vv
    # fourth sweep: axis is 'Y' or 'y'
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                saxis = self.fi.variables[vv].axis
            except AttributeError:
                saxis = ''
            if saxis.lower() == 'y':
                self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                saxis = self.fi.variables[vv].axis
            except AttributeError:
                saxis = ''
            if saxis.lower() == 'x':
                self.lonvar = vv
    # fifth sweep: same as first but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'latitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            try:
                sname = self.fi.variables[vv].standard_name
            except AttributeError:
                try:
                    sname = self.fi.variables[vv].long_name
                except AttributeError:
                    sname = self.fi.variables[vv].name
            if sname.lower() == 'longitude':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    # sixth sweep: same as second but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lat':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            if sname[0:3].lower() == 'lon':
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    # seventh sweep: same as third but units can be "degrees"
    if not self.latvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lat') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.latvar = vv
    if not self.lonvar:
        for vv in self.fi.variables:
            sname = self.fi.variables[vv].name
            sname = sname.lower()
            if sname.find('lon') >= 0:
                try:
                    sunit = self.fi.variables[vv].units
                except AttributeError:
                    sunit = ''
                if sunit.lower() == 'degrees':
                    self.lonvar = vv
    #
    # determine lat/lon dimensions
    self.latdim = ''
    self.londim = ''
    if self.latvar:
        latshape = self.fi.variables[self.latvar].shape
        if (len(latshape) < 1) or (len(latshape) > 2):
            estr  = 'Something went wrong determining lat/lon:'
            estr += ' latitude variable is not 1D or 2D.'
            print(estr)
            estr = 'latitude variable with dimensions:'
            ldim = self.fi.variables[self.latvar].dimensions
            print(estr, self.latvar, ldim)
            self.latvar = ''
        else:
            self.latdim = self.fi.variables[self.latvar].dimensions[0]
    if self.lonvar:
        lonshape = self.fi.variables[self.lonvar].shape
        if len(lonshape) == 1:
            self.londim = self.fi.variables[self.lonvar].dimensions[0]
        elif len(lonshape) == 2:
            self.londim = self.fi.variables[self.lonvar].dimensions[1]
        else:
            estr  = 'Something went wrong determining lat/lon:'
            estr += ' longitude variable is not 1D or 2D.'
            print(estr)
            estr = 'longitude variable with dimensions:'
            ldim = self.fi.variables[self.lonvar].dimensions
            print(estr, self.lonvar, ldim)
            self.lonvar = ''
    #
    # add units to lat/lon name
    if self.latvar:
        idim = tuple(zip_dim_name_length(self.fi.variables[self.latvar]))
        self.latvar = self.latvar + ' ' + str(idim)
    if self.lonvar:
        idim = tuple(zip_dim_name_length(self.fi.variables[self.lonvar]))
        self.lonvar = self.lonvar + ' ' + str(idim)


#
# Get list of missing values
#

def get_miss(self, x):
    """
    Get list of missing values, i.e. self.miss, x._FillValue,
    x.missing_value, and from netcdf4.default_fillvals.

    Parameters
    ----------
    self : class
        ncvue class
    x : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    list
        List with missing values self.miss, x._FillValue,
        x.missing_value if present,  and from netcdf4.default_fillvals

    Examples
    --------
    >>> x = fi.variables['time']
    >>> miss = get_miss(self, x)

    """
    try:
        out = [ncfill[x.dtype]]
    except KeyError:
        out = []
    try:
        if x.dtype != np.dtype('<M8[ms]'):
            out += [self.miss]
    except AttributeError:
        pass
    try:
        out += [x._FillValue]
    except AttributeError:
        pass
    try:
        out += [x.missing_value]
    except AttributeError:
        pass
    return out


#
# Get slice and set missing value
#

def get_slice_miss(self, dimspins, x):
    """
    Convenience method to get list of missing values (get_miss),
    choose slice of array (get_slice), and set missing values to
    NaN or np.datetime64('NaT') in slice (set_miss).

    Parameters
    ----------
    self : class
        ncvue class
    dimspins : list
        List of tk.Spinbox widgets of dimensions
    x : netCDF4._netCDF4.Variable
        netcdf variable

    Returns
    -------
    ndarray
        Extracted array slice with missing values set to
        np.NaN or np.datetime64('NaT')

    Examples
    --------
    >>> x = fi.variables['time']
    >>> xx = get_slice_miss(self, x)

    """
    miss = get_miss(self, x)
    xx   = get_slice(dimspins, x).squeeze()
    xx   = set_miss(miss, xx)
    # catch variables that have only one string or similar
    try:
        sx = xx.shape[0]
    except IndexError:
        xx = np.array([np.nan])
    return xx


#
# Set dimensions
#

# set_dim_lat/lon/var could also be methods of ncvMap.
# Leave them here for their concordance with set_dim_x/y/z

def set_dim_lat(self):
    """
    Set spinboxes of latitude-dimensions.

    Set labels and value lists. Select 'all' for all dimension.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of latitude-dimensions set.

    Examples
    --------
    >>> set_dim_lat(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.latd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.latdlblval[i].set(str(i))
        self.latdtip[i].set("")
    lat = self.lat.get()
    if lat != '':
        # set real dimensions
        vl = vardim2var(lat)
        if vl == self.tname:
            vl = self.tvar
        ll = self.fi.variables[vl]
        for i in range(ll.ndim):
            ww = max(4, int(np.ceil(np.log10(ll.shape[i]))))
            self.latd[i].config(values=spinbox_values(ll.shape[i]), width=ww,
                                state=tk.NORMAL)
            if (ll.shape[i] > 1):
                self.latdval[i].set('all')
            else:
                self.latdval[i].set(0)
            self.latdlblval[i].set(ll.dimensions[i])
            if ll.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    ll.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.latdtip[i].set(tstr)


def set_dim_lon(self):
    """
    Set spinboxes of longitude-dimensions.

    Set labels and value lists. Select 'all' for all dimension.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of longitude-dimensions set.

    Examples
    --------
    >>> set_dim_lon(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.lond[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.londlblval[i].set(str(i))
        self.londtip[i].set("")
    lon = self.lon.get()
    if lon != '':
        # set real dimensions
        vl = vardim2var(lon)
        if vl == self.tname:
            vl = self.tvar
        ll = self.fi.variables[vl]
        for i in range(ll.ndim):
            ww = max(4, int(np.ceil(np.log10(ll.shape[i]))))
            self.lond[i].config(values=spinbox_values(ll.shape[i]), width=ww,
                                state=tk.NORMAL)
            if (ll.shape[i] > 1):
                self.londval[i].set('all')
            else:
                self.londval[i].set(0)
            self.londlblval[i].set(ll.dimensions[i])
            if ll.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    ll.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.londtip[i].set(tstr)


def set_dim_var(self):
    """
    Set spinboxes of varable-dimensions in Map panel.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the first two limited dimensions and select
    0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of variable-dimensions set.

    Examples
    --------
    >>> set_dim_var(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.vd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.vdlblval[i].set(str(i))
        self.vdtip[i].set("")
    v = self.v.get()
    if v != '':
        # set real dimensions
        vz = vardim2var(v)
        if vz == self.tname:
            vz = self.tvar
        vv = self.fi.variables[vz]
        nall = 0
        if self.latdim:
            if self.latdim in vv.dimensions:
                i = vv.dimensions.index(self.latdim)
                ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
                self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                                  state=tk.NORMAL)
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(vv.dimensions[i])
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
        if self.londim:
            if self.londim in vv.dimensions:
                i = vv.dimensions.index(self.londim)
                ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
                self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                                  state=tk.NORMAL)
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(vv.dimensions[i])
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
        for i in range(vv.ndim):
            ww = max(5, int(np.ceil(np.log10(vv.shape[i]))))  # 5~median
            self.vd[i].config(values=spinbox_values(vv.shape[i]), width=ww,
                              state=tk.NORMAL)
            if ((vv.dimensions[i] != self.latdim) and
                (vv.dimensions[i] != self.londim) and
                (vv.dimensions[i] != self.dunlim) and
                (nall <= 1) and (vv.shape[i] > 1)):
                nall += 1
                self.vdval[i].set('all')
                self.vdlblval[i].set(vv.dimensions[i])
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)
            elif ((vv.dimensions[i] != self.latdim) and
                  (vv.dimensions[i] != self.londim)):
                self.vdval[i].set(0)
                self.vdlblval[i].set(vv.dimensions[i])
                if vv.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        vv.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.vdtip[i].set(tstr)


def set_dim_x(self):
    """
    Set spinboxes of x-dimensions.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of x-dimensions set.

    Examples
    --------
    >>> set_dim_x(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.xd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.xdlblval[i].set(str(i))
        self.xdtip[i].set("")
    x = self.x.get()
    if x != '':
        # set real dimensions
        vx = vardim2var(x)
        if vx == self.tname:
            vx = self.tvar
        xx = self.fi.variables[vx]
        nall = 0
        if self.dunlim in xx.dimensions:
            i = xx.dimensions.index(self.dunlim)
            ww = max(5, int(np.ceil(np.log10(xx.shape[i]))))  # 5~median
            self.xd[i].config(values=spinbox_values(xx.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.xdval[i].set('all')
            self.xdlblval[i].set(xx.dimensions[i])
            if xx.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    xx.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.xdtip[i].set(tstr)
        for i in range(xx.ndim):
            if xx.dimensions[i] != self.dunlim:
                ww = max(5, int(np.ceil(np.log10(xx.shape[i]))))
                self.xd[i].config(values=spinbox_values(xx.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall == 0) and (xx.shape[i] > 1):
                    nall += 1
                    self.xdval[i].set('all')
                else:
                    self.xdval[i].set(0)
                self.xdlblval[i].set(xx.dimensions[i])
                if xx.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        xx.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.xdtip[i].set(tstr)


def set_dim_y(self):
    """
    Set spinboxes of y-dimensions of the left-hand-side (lhs).

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of lhs y-dimensions set.

    Examples
    --------
    >>> set_dim_y(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.yd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.ydlblval[i].set(str(i))
        self.ydtip[i].set("")
    y = self.y.get()
    if y != '':
        # set real dimensions
        vy = vardim2var(y)
        if vy == self.tname:
            vy = self.tvar
        yy = self.fi.variables[vy]
        nall = 0
        if self.dunlim in yy.dimensions:
            i = yy.dimensions.index(self.dunlim)
            ww = max(5, int(np.ceil(np.log10(yy.shape[i]))))  # 5~median
            self.yd[i].config(values=spinbox_values(yy.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.ydval[i].set('all')
            self.ydlblval[i].set(yy.dimensions[i])
            if yy.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    yy.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.ydtip[i].set(tstr)
        for i in range(yy.ndim):
            if yy.dimensions[i] != self.dunlim:
                ww = max(5, int(np.ceil(np.log10(yy.shape[i]))))
                self.yd[i].config(values=spinbox_values(yy.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall == 0) and (yy.shape[i] > 1):
                    nall += 1
                    self.ydval[i].set('all')
                else:
                    self.ydval[i].set(0)
                self.ydlblval[i].set(yy.dimensions[i])
                if yy.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        yy.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.ydtip[i].set(tstr)


def set_dim_y2(self):
    """
    Set spinboxes of y2-dimensions of the right-hand-side (rhs).

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, and select 0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of rhs y-dimensions set.

    Examples
    --------
    >>> set_dim_y2(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.y2d[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.y2dlblval[i].set(str(i))
        self.y2dtip[i].set("")
    y2 = self.y2.get()
    if y2 != '':
        # set real dimensions
        vy2 = vardim2var(y2)
        if vy2 == self.tname:
            vy2 = self.tvar
        yy2 = self.fi.variables[vy2]
        nall = 0
        if self.dunlim in yy2.dimensions:
            i = yy2.dimensions.index(self.dunlim)
            ww = max(5, int(np.ceil(np.log10(yy2.shape[i]))))  # 5~median
            self.y2d[i].config(values=spinbox_values(yy2.shape[i]), width=ww,
                               state=tk.NORMAL)
            nall += 1
            self.y2dval[i].set('all')
            self.y2dlblval[i].set(yy2.dimensions[i])
            if yy2.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    yy2.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.y2dtip[i].set(tstr)
        for i in range(yy2.ndim):
            if yy2.dimensions[i] != self.dunlim:
                ww = max(5, int(np.ceil(np.log10(yy2.shape[i]))))
                self.y2d[i].config(values=spinbox_values(yy2.shape[i]),
                                   width=ww, state=tk.NORMAL)
                if (nall == 0) and (yy2.shape[i] > 1):
                    nall += 1
                    self.y2dval[i].set('all')
                else:
                    self.y2dval[i].set(0)
                self.y2dlblval[i].set(yy2.dimensions[i])
                if yy2.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        yy2.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.y2dtip[i].set(tstr)


def set_dim_z(self):
    """
    Set spinboxes of z-dimensions.

    Set labels and value lists, including 'all' to select all entries,
    as well as 'mean', 'std', etc. for common operations on the axis.

    Select 'all' for the unlimited dimension if it exists, otherwise
    for the first dimension, as well as for a second dimension and select
    0 for all other dimensions.

    Parameters
    ----------
    self : class
        ncvue class

    Returns
    -------
    None
        Labels and values of spinboxes of z-dimensions set.

    Examples
    --------
    >>> set_dim_z(self)

    """
    # reset dimensions
    for i in range(self.maxdim):
        self.zd[i].config(values=(0,), width=1, state=tk.DISABLED)
        self.zdlblval[i].set(str(i))
        self.zdtip[i].set("")
    z = self.z.get()
    if z != '':
        # set real dimensions
        vz = vardim2var(z)
        if vz == self.tname:
            vz = self.tvar
        zz = self.fi.variables[vz]
        nall = 0
        if self.dunlim in zz.dimensions:
            i = zz.dimensions.index(self.dunlim)
            ww = max(5, int(np.ceil(np.log10(zz.shape[i]))))  # 5~median
            self.zd[i].config(values=spinbox_values(zz.shape[i]), width=ww,
                              state=tk.NORMAL)
            nall += 1
            self.zdval[i].set('all')
            self.zdlblval[i].set(zz.dimensions[i])
            if zz.shape[i] > 1:
                tstr  = "Specific dimension value: 0-{:d}\n".format(
                    zz.shape[i]-1)
                tstr += "or arithmetic operation on axis:\n"
                tstr += "  " + ", ".join(DIMMETHODS)
            else:
                tstr = "Single dimension: 0"
            self.zdtip[i].set(tstr)
        for i in range(zz.ndim):
            if zz.dimensions[i] != self.dunlim:
                ww = max(5, int(np.ceil(np.log10(zz.shape[i]))))
                self.zd[i].config(values=spinbox_values(zz.shape[i]), width=ww,
                                  state=tk.NORMAL)
                if (nall <= 1) and (zz.shape[i] > 1):
                    nall += 1
                    self.zdval[i].set('all')
                else:
                    self.zdval[i].set(0)
                self.zdlblval[i].set(zz.dimensions[i])
                if zz.shape[i] > 1:
                    tstr  = "Specific dimension value: 0-{:d}\n".format(
                        zz.shape[i]-1)
                    tstr += "or arithmetic operation on axis:\n"
                    tstr += "  " + ", ".join(DIMMETHODS)
                else:
                    tstr = "Single dimension: 0"
                self.zdtip[i].set(tstr)



#!/usr/bin/env python
"""
Widget functions for ncvue.

Convenience functions for adding Tkinter widgets.

This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2021 Matthias Cuntz - mc (at) macu (dot) de
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   Tooltip
   add_checkbutton
   add_combobox
   add_entry
   add_imagemenu
   add_menu
   add_scale
   add_spinbox
   add_tooltip

History
    * Written Nov-Dec 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added tooltips to all widgets with class Tooltip,
      Jan 2021, Matthias Cuntz
    * Added add_tooltip widget, Jan 2021, Matthias Cuntz
    * add_spinbox returns also label widget, Jan 2021, Matthias Cuntz

"""
from __future__ import absolute_import, division, print_function
import tkinter as tk
try:
    import tkinter.ttk as ttk
except Exception:
    import sys
    print('Using the themed widget set introduced in Tk 8.5.')
    print('Try to use mcview.py, which uses wxpython instead.')
    sys.exit()
from idlelib.tooltip import Hovertip


__all__ = ['Tooltip',
           'add_checkbutton', 'add_combobox', 'add_entry',
           'add_imagemenu', 'add_menu', 'add_scale', 'add_spinbox',
           'add_tooltip']


class Tooltip(Hovertip):
    """
    A tooltip that pops up when a mouse hovers over an anchor widget.

    This is a copy of the class Hovertip of Python's idlelib/tooltip.py.
    In addition, it sets the foreground colour to see the tip also in
    macOS dark mode, and displays a textvariable rather than simple text
    so one can change the tip during run time.

    """
    def __init__(self, anchor_widget, text, hover_delay=1000):
        """
        Create a text tooltip with a mouse hover delay.
        anchor_widget: the widget next to which the tooltip will be shown
        text: tk.StringVar with text to display
        hover_delay: time to delay before showing the tooltip, in milliseconds
        Note that a widget will only be shown when showtip() is called,
        e.g. after hovering over the anchor widget with the mouse for enough
        time.

        """
        super(Tooltip, self).__init__(anchor_widget, text,
                                      hover_delay=hover_delay)

    def showcontents(self):
        # light yellow = #ffffe0
        label = tk.Label(self.tipwindow, textvariable=self.text,
                         background="#ffffe0", foreground="#000000",
                         justify=tk.LEFT, relief=tk.FLAT, borderwidth=0,
                         padx=1, pady=1)
        label.pack()


def add_checkbutton(frame, label="", value=False, command=None, tooltip="",
                    **kwargs):
    """
    Add a left-aligned ttk.Checkbutton.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears on the checkbutton (default: "")
    value : bool, optional
        Initial state of the checkbutton (default: False)
    command : function, optional
        Function to be called whenever the state of the
        checkbutton changes (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the checkbutton (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to ttk.Checkbutton

    Returns
    -------
    tk.StringVar, tk.BooleanVar
        variable for the text on the checkbutton,
        control variable tracking current state of checkbutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = ttk.Frame(self)
    >>> self.rowzxy.pack(side=tk.TOP, fill=tk.X)
    >>> self.inv_xlbl, self.inv_x = add_checkbutton(
    ...     self.rowzxy, label="invert x", value=False, command=self.checked)

    """
    check_label = tk.StringVar()
    check_label.set(label)
    bvar = tk.BooleanVar(value=value)
    cb = ttk.Checkbutton(frame, variable=bvar, textvariable=check_label,
                         command=command, **kwargs)
    cb.pack(side=tk.LEFT, padx=3)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        cbtip = Tooltip(cb, ttip)
        return check_label, bvar, ttip
    else:
        return check_label, bvar


def add_combobox(frame, label="", values=[], command=None, tooltip="",
                 **kwargs):
    """
    Add a left-aligned ttk.Combobox with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the combobox (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    command : function, optional
        Handler function to be bound to the combobox for the event
        <<ComboboxSelected>> (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the combobox (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to ttk.Combobox

    Returns
    -------
    tk.StringVar, ttk.Combobox
        variable for the text before the combobox, combobox widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = ttk.Frame(self)
    >>> self.rowzxy.pack(side=tk.TOP, fill=tk.X)
    >>> self.xlbl, self.x = add_combobox(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    width = kwargs.pop('width', 25)
    cb_label = tk.StringVar()
    cb_label.set(label)
    label = ttk.Label(frame, textvariable=cb_label)
    label.pack(side=tk.LEFT)
    cb = ttk.Combobox(frame, values=values, width=width, **kwargs)
    # long = len(max(values, key=len))
    # cb.configure(width=(max(20, long//2)))
    if command is not None:
        cb.bind("<<ComboboxSelected>>", command)
    cb.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        cbtip = Tooltip(cb, ttip)
        return cb_label, cb, ttip
    else:
        return cb_label, cb


def add_entry(frame, label="", text="", command=None, tooltip="", **kwargs):
    """
    Add a left-aligned ttk.Entry with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the entry (default: "")
    text : str, optional
        Initial text in the entry area (default: "")
    command : function, optional
        Handler function to be bound to the entry for the events
        <Return>, '<Key-Return>', <KP_Enter>, and '<FocusOut>' (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the entry (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to ttk.Entry

    Returns
    -------
    tk.StringVar, tk.StringVar
        variable for the text before the entry,
        variable for the text in the entry area
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowxyopt = ttk.Frame(self)
    >>> self.rowxyopt.pack(side=tk.TOP, fill=tk.X)
    >>> self.lslbl, self.ls = add_entry(
    ...     self.rowxyopt, label="ls", text='-',
    ...     width=4, command=self.selected_y)

    """
    entry_label = tk.StringVar()
    entry_label.set(label)
    label = ttk.Label(frame, textvariable=entry_label)
    label.pack(side=tk.LEFT)
    entry_text = tk.StringVar()
    entry_text.set(text)
    entry = ttk.Entry(frame, textvariable=entry_text, **kwargs)
    if command is not None:
        entry.bind('<Return>', command)      # return
        entry.bind('<Key-Return>', command)  # return
        entry.bind('<KP_Enter>', command)    # return of numeric keypad
        entry.bind('<FocusOut>', command)    # tab or click
    entry.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        etip = Tooltip(entry, ttip)
        return entry_label, entry_text, ttip
    else:
        return entry_label, entry_text


def add_imagemenu(frame, label="", values=[], images=[], command=None,
                  tooltip="", **kwargs):
    """
    Add a left-aligned menu with menubuttons having text and images
    with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the menu (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    images : list of str, optional
        The images before the choices that will appear in the drop-down menu
        (default: [])
    command : function, optional
        Handler function to be called if new menu entry chosen (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the menu (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to the main ttk.Menubutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Returns
    -------
    tk.StringVar, ttk.Menubutton
        variable for the text before the menu, main tt.Menubutton widget

    Examples
    --------
    >>> self.rowcmap = ttk.Frame(self)
    >>> self.rowcmap.pack(side=tk.TOP, fill=tk.X)
    >>> self.cmaplbl, self.cmap = add_imagemenu(
    ...     self.rowcmap, label="cmap", values=self.cmaps,
    ...     images=self.imaps, command=self.selected_cmap)
    >>> self.cmap['text']  = 'RdYlBu'
    >>> self.cmap['image'] = self.imaps[self.cmaps.index('RdYlBu')]

    """
    from functools import partial
    estr  = 'Same number of values and images needed for add_imagemenu.'
    estr += ' values (' + str(len(values)) + '): ' + str(values)
    estr += ', images (' + str(len(images)) + '): ' + str(images)
    assert len(values) == len(images), estr
    mb_label = tk.StringVar()
    mb_label.set(label)
    label = ttk.Label(frame, textvariable=mb_label)
    label.pack(side=tk.LEFT)
    mb = ttk.Menubutton(frame, image=images[0], text=values[0],
                        compound=tk.LEFT)
    sb = tk.Menu(mb, tearoff=False)
    mb.config(menu=sb)
    for i, v in enumerate(values):
        sb.add_command(label=v, image=images[i], compound=tk.LEFT,
                       command=partial(command, v))
    mb.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        mbtip = Tooltip(mb, ttip)
        return mb_label, mb, ttip
    else:
        return mb_label, mb


def add_menu(frame, label="", values=[], command=None, tooltip="", **kwargs):
    """
    Add a left-aligned menu with menubuttons with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the menu (default: "")
    values : list of str, optional
        The choices that will appear in the drop-down menu (default: [])
    command : function, optional
        Handler function to be called if new menu entry chosen (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the menu (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to the main ttk.Menubutton
    tk.StringVar
        variable for the text of the tooltip, if given.

    Returns
    -------
    tk.StringVar, ttk.Menubutton
        variable for the text before the menu, main tt.Menubutton widget

    Examples
    --------
    >>> self.rowzxy = ttk.Frame(self)
    >>> self.rowzxy.pack(side=tk.TOP, fill=tk.X)
    >>> self.xlbl, self.x = add_combobox(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    from functools import partial
    mb_label = tk.StringVar()
    mb_label.set(label)
    label = ttk.Label(frame, textvariable=mb_label)
    label.pack(side=tk.LEFT)
    mb = ttk.Menubutton(frame, text=values[0], compound=tk.LEFT)
    sb = tk.Menu(mb, tearoff=False)
    mb.config(menu=sb)
    for i, v in enumerate(values):
        sb.add_command(label=v, compound=tk.LEFT,
                       command=partial(command, v))
    mb.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        mbtip = Tooltip(mb, ttip)
        return mb_label, mb, ttip
    else:
        return mb_label, mb


def add_scale(frame, label="", ini=0, tooltip="", **kwargs):
    """
    Add a left-aligned ttk.Scale with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the scale (default: "")
    ini : float, optional
        Initial value of scale (default: 0)
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the scale (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to ttk.Scale

    Returns
    -------
    tk.StringVar, tk.DoubleVar, ttk.Scale
        variable for the text before the scale,
        value of scale,
        scale widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowzxy = ttk.Frame(self)
    >>> self.rowzxy.pack(side=tk.TOP, fill=tk.X)
    >>> self.xlbl, self.x = add_scale(
    ...     self.rowzxy, label="x", values=columns, command=self.selected)

    """
    s_label = tk.StringVar()
    s_label.set(label)
    label = ttk.Label(frame, textvariable=s_label)
    label.pack(side=tk.LEFT)
    s_val = tk.DoubleVar()
    s_val.set(ini)
    s = ttk.Scale(frame, variable=s_val, **kwargs)
    s.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        stip = Tooltip(s, ttip)
        return s_label, s_val, s, ttip
    else:
        return s_label, s_val, s


def add_spinbox(frame, label="", values=[], command=None, tooltip="",
                **kwargs):
    """
    Add a left-aligned tk.Spinbox with a ttk.Label before.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    label : str, optional
        Text that appears in front of the spinbox (default: "")
    values : list of str, optional
        The list of choices on the spinbox (default: [])
    command : function, optional
        Handler function bound to
        <Return>, '<Key-Return>', <KP_Enter>, and '<FocusOut>' (default: None).
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the spinbox (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to tk.Spinbox

    Returns
    -------
    tk.StringVar, ttk.Label, tk.StringVar, tk.Spinbox
        variable for the text before the spinbox,
        label widget,
        variable for the text in the spinbox area,
        spinbox widget
    tk.StringVar
        variable for the text of the tooltip, if given.

    Examples
    --------
    >>> self.rowlev = ttk.Frame(self)
    >>> self.rowlev.pack(side=tk.TOP, fill=tk.X)
    >>> self.dlval, self.dl, self.dval, self.d = add_spinbox(
    ...     self.rowlev, label="dim", values=range(0,10),
    ...     command=self.spinned)

    """
    width = kwargs.pop('width', 1)
    sbl_val = tk.StringVar()
    sbl_val.set(label)
    sbl = ttk.Label(frame, textvariable=sbl_val)
    sbl.pack(side=tk.LEFT)
    sb_val = tk.StringVar()
    if len(values) > 0:
        sb_val.set(str(values[0]))
    sb = tk.Spinbox(frame, values=values, command=command, width=width,
                    textvariable=sb_val, **kwargs)
    if command is not None:
        sb.bind('<Return>', command)      # return
        sb.bind('<Key-Return>', command)  # return
        sb.bind('<KP_Enter>', command)    # return of numeric keypad
        sb.bind('<FocusOut>', command)    # tab or click
    sb.pack(side=tk.LEFT)
    if tooltip:
        ttip = tk.StringVar()
        ttip.set(tooltip)
        sbtip = Tooltip(sb, ttip)
        return sbl_val, sbl, sb_val, sb, ttip
    else:
        return sbl_val, sbl, sb_val, sb


def add_tooltip(frame, tooltip="", **kwargs):
    """
    Add a writeable tooltip to a widget using the class Tooltip.

    Parameters
    ----------
    frame : tk widget
        Parent widget
    tooltip : str, optional
        Tooltip appearing after one second when hovering over
        the parent widget (default: "" = no tooltip)
    **kwargs : option=value pairs, optional
        All other options will be passed to tk.StringVar

    Returns
    -------
    tk.StringVar
        variable for the text of the tooltip.

    Examples
    --------
    >>> self.rowlev = ttk.Frame(self)
    >>> self.rowlev.pack(side=tk.TOP, fill=tk.X)
    >>> self.dlbl, self.dval, self.d = add_spinbox(
    ...     self.rowlev, label="dim", values=range(0,10),
    ...     command=self.spinned)
    >>> self.dtip = add_tooltip(self.d, 'Dimension')

    """
    ttip = tk.StringVar()
    ttip.set(tooltip)
    htip = Tooltip(frame, ttip)
    return ttip
